/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.algorithm.maths;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

/** 
 * Combinatorics: maths class. Based mainly on the computation of sets
 * of combinations without repetition.
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 *
 * @version 0.1 (2011-01-09)
 */
public class Combinatorics {
	
	private static Logger LOGGER = Logger.getLogger(Combinatorics.class);
	
	/////////////////////////////////////////////////////////////////////////////////////
	///////////////// BASIC COMBINATORICS: FACTORIAL, COMBINATION, ETC. /////////////////
	/////////////////////////////////////////////////////////////////////////////////////
	
	/** 
	 * Factorial.
	 * 
	 * @param n an integer
	 * @return n! (n factorial)
	 */
	public static double factorial(int n) {
		if (n == 0) return 1;
		return n * (factorial(n - 1));
	}
	
	
	/** 
	 * Combination without repetition C(n,k).
	 * 
	 * @param n an integer
	 * @param k
	 * @return C(n,k)
	 */
	public static double combination(int n, int k) {
		return factorial(n) / (factorial(k) * factorial(n - k));
	}


	/////////////////////////////////////////////////////////////////////////////////////
	////////////////////////// COMBINATION WITHOUT REPETITION ///////////////////////////
	//////////////// COMPUTATION OF ALL POSSIBLE FOR N BETWEEN 0 AND N-1 ////////////////
	/////////////////////////////////////////////////////////////////////////////////////

	private static List<List<Integer>> list = new ArrayList<List<Integer>>();
	
	
	/** 
	 * Compute and return the list of the lists of the set of all possible combinations
	 * without repetition for the numbers between 0 and n-1. For example, return the
	 * following combinations if n = 3, i.e. using the set {0,1,2,}: {0}, {1}, {2}, {0,1},
	 * {0,2}, {1,2}, {0,1,2}.
	 * ATTENTION: Works fine for a value of n until 24. More than 24, it requires too much
	 * memory.
	 * 
	 * @param n
	 * @return the list of the lists of the set of all possible combinations without repetition
	 * for the numbers between 0 and n-1
	 */
	public static List<List<Integer>> combination(int n) {
		list = new ArrayList<List<Integer>>();
		if (n == 0) return list;
		for (int i = 0 ; i <= n ; i++) {
			int[] combinations = new int[i];
			combination(combinations, 0, 0, i, n);
		}
		return list;
	}
	
	
	/** 
	 * (Sub-method of the method combination)
	 * Compute the list of the lists of the set of all possible combinations without
	 * repetition for the numbers between 0 and n-1. For example, return the following
	 * following combinations if n = 3, i.e. using the set {0,1,2,}: {0}, {1}, {2}, {0,1},
	 * {0,2}, {1,2}, {0,1,2}.
	 * 
	 * @param combinations
	 * @param position
	 * @param nextInt
	 * @param k
	 * @param n
	 */
	private static void combination(int[] combinations, int position, int nextInt, int k, int n) {
		if (position == k) {
			List<Integer> l = new ArrayList<Integer>();
			for (int i = 0 ; i < combinations.length ; i++) {
				l.add(combinations[i]);
			}
			if (! l.isEmpty()) {
				list.add(l);			
			}
			return;
		}
		for (int i = nextInt ; i < n ; i++) {
			combinations[position] = i;
			combination(combinations, position + 1, i + 1, k, n);
		}
	}
	
	
	/** 
	 * Show a combination.
	 * 
	 * @param combination
	 */
	public static void showCombination(List<List<Integer>> combination) {
		LOGGER.info("COMBINATION");
		for (List<Integer> l : combination) {
			LOGGER.info(l);
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////
	////////////////////////// COMBINATION WITHOUT REPETITION ///////////////////////////
	//////////////////////////          STRING PROCESS        ///////////////////////////	
	/////////////////////////////////////////////////////////////////////////////////////

	/** 
	 * (Based on String) Compute and return the list of the lists of the set of all possible
	 * combinations without repetition for the numbers between 0 and n-1. For example, return
	 * the following combinations if n = 3, i.e. using the set {0,1,2,}: {0}, {1}, {2}, {0,1},
	 * {0,2}, {1,2}, {0,1,2}. 
	 * ATTENTION: Works only until 10 for the value of n (due to the fact that this method is
	 * based on String elements).
	 * 
	 * @param n
	 * @return the list of the lists of the set of all possible combinations without repetition
	 * for the numbers between 0 and n-1
	 */
	public static List<List<Integer>> combinationString(int n) {
		if (n == 0) return list;
		String s = "";
		for (int i = 0; i < n ; i++) {
			s = s.concat(new Integer(i).toString());
		}
		combinationString("", s);
		return list;
	}

	
	/** 
	 * Sub-method of the method combinationString.
	 * 
	 * @param bootstrap
	 * @param s
	 */
	private static void combinationString(String bootstrap, String s) {
		List<Integer> l = new ArrayList<Integer>();
		for (int i = 0; i < bootstrap.length() ; i++) {
			l.add(new Integer(new Character(bootstrap.charAt(i)).toString()));
		}
		if (! l.isEmpty()) {
			list.add(l);
		}
		for (int i = 0; i < s.length() ; i++) {
			combinationString(bootstrap + s.charAt(i), s.substring(i + 1));
		}
	}
	
	
	/////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////
	
	/** 
	 * Example through a main method.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		
		// To know how many cases there are
		int n = 16;
		int a = 0;
		for (int i = 1 ; i <= n ; i++) {
			a += combination(16,i);
		}
		
		LOGGER.info("Number of cominations: " + a);
		
		int limit = 14;
		for (int i = 0 ; i <= limit ; i++) {
			LOGGER.info("fact(" + i + ")=" + factorial(i));			
		}
		
		List<List<Integer>> combinations = combination(14);
		showCombination(combinations);
	}
	
}