/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.geometry;

import com.vividsolutions.jts.geom.Point;

/**
 * Geometric methods to compute angles.
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 * 
 * @version 0.1 (2011-01-09)
 */
public class AngleUtil {

	private final static int DEGRE = 0;
	private final static int GRADE = 2;
	private final static int ZERO_PI = 0;
	
	// unuseful variable except to provide a context for the [0 ; Pi/2[ interval
	@SuppressWarnings("unused")
	private final static int ZERO_PI_2 = 1;

	
	/** 
	 * Compute the angle between 2 {@link Point}s (between 0 and PI or between O and PI/2)
	 * comparatively to the X axis.
	 * 
	 * @param point first point of the angle
	 * @param point2 point to process
	 * @param unit unit in degree if 0, grade if 2
	 * @param interval inteval [ 0 ; PI [ if 0, [ 0 ; PI/2 [ if 1,
	 * 		  otherwise the interval [ 0 ; PI/2 [ is the default choice
	 * @param accuracy true to keep two numbers after the dot, false to keep all numbers
	 * @return the computed angle
	 */
	public static double angle(Point point, Point point2, int unit, int interval, boolean accuracy) {
		double x = point2.getX() - point.getX();
		double y = point2.getY() - point.getY();
		double angle = Math.atan2(y,x);

		// [O ; Pi[
		if (interval == AngleUtil.ZERO_PI) {
			if (angle < 0) angle = angle + Math.PI;
			if (angle == Math.PI) angle = 0.0;
		} else { // [O ; Pi/2[ (by choice or by default)
			if (angle < 0) angle = angle + Math.PI;
			angle = angle % (Math.PI / 2.0);
			if (angle == Math.PI/2.0) angle = 0.0;
		}
		
		// Convert into the chosen unit
		if (unit == AngleUtil.DEGRE) {
			angle = Math.toDegrees(angle);
		} else if (unit == AngleUtil.GRADE) {
			angle = angle * 200.0 / Math.PI;
		}
		
		if (accuracy) {
			angle = new Long(Math.round(angle * 100)).doubleValue() / 100.0;	
		}
		
		return angle;
	}

}