/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.geometry;

import java.util.ArrayList;
import java.util.List;

import com.vividsolutions.jts.algorithm.Angle;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.CoordinateSequence;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.impl.CoordinateArraySequenceFactory;

/**  
 * Geometric methods to manipulate geometric features with {@link LineString}
 * and {@link MultiLineString} types (cleaning methods only at the moment).
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 *
 * @version 0.1 (2011-01-09)
 */
public class LineStringUtil {
	
	/** 
	 * Clean a {@link LineString} by removing the unuseful points (in regard to the
	 * chosen threshold) of a {@link LineString}, i.e. the points which are in the 
	 * alignment between the previous point and the following one. The alignment
	 * is here defined thanks to a given threshold comparatively to Pi (or 180°).
	 * The threshold has to be expressed in degrees.
	 * 
	 * @param line linestring to clean
	 * @param threshold threshold in degrees to remove points which are aligned
	 * with an angle of 180 +/- threshold in the linestring
	 * @return a new and cleaned linestring
	 */
	public static LineString cleaningLineShape(LineString line, double threshold) {

		// Initialise a Geometryfactory 
		GeometryFactory factory = new GeometryFactory(line.getPrecisionModel(), line.getSRID());
		
		// Coordinates of the LineString
		Coordinate[] coordT = line.getCoordinates();
		
		// List of a coordinates list
		List<Coordinate> coords = new ArrayList<Coordinate>();
		for (int i = 0 ; i < coordT.length ; i++) {
			coords.add(coordT[i]);
		}
		
		// Dectect unuseful coordinates
		List<Integer> toRemove = new ArrayList<Integer>();
		for (int i = 0 ; i < coords.size() - 2 ; i++) {
			Coordinate c0 = coordT[i], 
						c1 = coordT[i + 1], 
						c2 = coordT[i + 2];
			
			double angle = Angle.interiorAngle(c0, c1, c2);
			
			if ( angle >= Math.PI - Math.toRadians(threshold)
					&& angle <= Math.PI + Math.toRadians(threshold)) {
				toRemove.add(0, i + 1);
			}
		}
		
		// Delete unuseful coordinates
		for (Integer i : toRemove) {
			coords.remove(i.intValue());
		}
		
		// Convert into an array and create the CoordinateSequence
		// associated to the simplified coordinates
		CoordinateSequence cs = CoordinateArraySequenceFactory.instance().create(coords.toArray(new Coordinate[coords.size()]));
				
		return new LineString(cs, factory);
	}
	
	
	/** 
	 * Clean a {@link MultiLineString} by removing the unuseful points (in regard to the
	 * chosen threshold) of a {@link MultiLineString}, i.e. the points which are in the 
	 * alignment between the previous point and the following one. The alignment
	 * is here defined thanks to a given threshold comparatively to Pi (or 180°).
	 * The threshold has to be expressed in degrees.
	 * 
	 * @param multiLine multilinestring to clean
	 * @param threshold threshold in degrees to remove points which are aligned
	 * with an angle of 180 +/- threshold in the multilinestring
	 * @return a new and cleaned multilinestring
	 */	
	public static MultiLineString cleaningMultiLineShape(MultiLineString multiLine, double threshold) {
		
		LineString[] lines = new LineString[multiLine.getNumGeometries()];
		
		// Use the cleaning method cleaningLineShape
		for (int i = 0 ; i < multiLine.getNumGeometries() ; i++) {
			LineString line = (LineString)multiLine.getGeometryN(i);
			lines[i] = cleaningLineShape(line, threshold);
		}
		
		// Initialise a Geometryfactory and return a new and clean multiline
		return new MultiLineString(lines, new GeometryFactory(multiLine.getPrecisionModel(), multiLine.getSRID()));
	}

}