/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.geometry;

import java.util.ArrayList;
import java.util.List;

import com.vividsolutions.jts.algorithm.Angle;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineSegment;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Polygon;

/**  
 * Geometric methods to manipulate geometric features with {@link Polygon}
 * and {@link MultiPolygon} types, and to compute indices.
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 * 
 * @version 0.1 (2011-01-09)
 */
public class PolygonUtil {
	
	//////////////////////////////////////////////////////////////////
	//////////// COMPACTEDNESS INDEX AND SURFACE DISTANCE ////////////
	//////////////////////////////////////////////////////////////////
	
	/** 
	 * Compute the compactedness index of a {@link Polygon}, i.e. Miller's index,
	 * 4.Pi.area / perimeter^2.
	 * 
	 * @param polygon polygon to process
	 * @return the compactedness index of the polygon
	 */
	public static double compactednessIndex(Polygon polygon) {
		if (polygon.getCoordinates().length < 4) return 0;
		double perimeter = polygon.getLength();
		if ( perimeter == 0 ) return 0;
		double area = polygon.getArea();
		return 4*(Math.PI)*area / Math.pow(perimeter,2);
	}

	
	/** 
	 * Compute the surface distance between two {@link Polygon}s.
	 * Surface distance is a number between 0 and 1, computed thanks to 
	 * the following formula: 1 - (area(intersection) / area(union)).
	 * 
	 * @param polygonA first polygon to process
	 * @param polygonB second polygon to process
	 * @return the surface distance between polygon A and polygon B
	 */
	public static double surfaceDistance(Polygon polygonA, Polygon polygonB) {        
		Geometry intersection = polygonA.intersection(polygonB);
		if ( intersection == null ) return 1;
		Geometry union = polygonA.union(polygonB);
		if ( union == null ) return 1;
		return 1 - intersection.getArea() / union.getArea();
	}

	
	/** 
	 * Compute the surface distance between two {@link Geometry}s.
	 * Surface distance is a number between 0 and 1, computed thanks to 
	 * the following formula: 1 - (area(intersection) / area(union)).
	 * 
	 * @param geometryA first geometry to process
	 * @param geometryB second geometry to process
	 * @return the surface distance between geometry A and geometry B
	 */
	public static double surfaceDistance(Geometry geometryA, Geometry geometryB) {
		Geometry intersection = geometryA.intersection(geometryB);
		if ( intersection == null ) return 1;
		Geometry union = geometryA.union(geometryB);
		if ( union == null ) return 1;
		return 1 - intersection.getArea() / union.getArea();
	}
	
	
	//////////////////////////////////////////////////////////////////
	////////////// ROTATION, MINIMUM BOUNDING RECTANGLE, /////////////
	/////////////       ELONGATION AND ORIENTATION       /////////////
	//////////////////////////////////////////////////////////////////
	
	/** 
	 * Compute the rotation of a {@link Polygon}.
	 * 
	 * @param polygon polygon to process
	 * @param center rotation center
	 * @param angle rotation angle
	 * @return the rotated polygon
	 */
	public static Polygon rotation(Polygon polygon, Coordinate center, double angle){
		double cos = Math.cos(angle);
		double sin = Math.sin(angle);
		
		// Exterior ring rotation
		Coordinate[] coord = polygon.getExteriorRing().getCoordinates();
		Coordinate[] coordRotated = new Coordinate[coord.length];
		for(int i = 0 ; i < coord.length ; i++) {
			double x = coord[i].x;
			double y = coord[i].y;
			coordRotated[i] = new Coordinate(
					center.x + cos * (x-center.x) - sin * (y-center.y),
					center.y + sin * (x-center.x) + cos * (y-center.y));
		}
		
		LinearRing shell = polygon.getFactory().createLinearRing(coordRotated);

		// Interior rings (holes) rotation
		LinearRing[] holes = new LinearRing[polygon.getNumInteriorRing()];
		for(int j = 0 ; j < polygon.getNumInteriorRing() ; j++) {
			Coordinate[] coordRing = polygon.getInteriorRingN(j).getCoordinates();
			Coordinate[] coordRingRotated = new Coordinate[coordRing.length];
			
			for(int i=0 ; i < coordRing.length ; i++) {
				double x = coordRing[i].x;
				double y = coordRing[i].y;
				coordRingRotated[i] = new Coordinate(
						center.x + cos * (x-center.x) - sin * (y-center.y),
						center.y + sin * (x-center.x) + cos * (y-center.y));
			}
			
			holes[j] = polygon.getFactory().createLinearRing(coordRing);
		}
		return polygon.getFactory().createPolygon(shell, holes);
	}
	
	
	/** 
	 * Compute the minimum bounding rectangle (MBR) of a {@link Polygon}.
	 * 
	 * @param polygon polygon to process
	 * @return the mimimum bounding rectangle, or null if the MBR can not be computed
	 */
	public static Polygon MBR(Polygon polygon){
		
		Geometry cHull = polygon.convexHull();
		
		// Return null if the convex hull geometry is not a Polygon
		if (!(cHull instanceof Polygon)) return null;
		Polygon convexHull = (Polygon)cHull;

		Coordinate[] coord = convexHull.getExteriorRing().getCoordinates();
		Coordinate center = polygon.getCentroid().getCoordinate();

		double areaMin = Double.MAX_VALUE;
		double angleForRotation = 0.0;
		Polygon mbr = null;
		for(int i = 0 ; i < coord.length - 1 ; i++) {
			
			// Rotation of the convex hull
			double angle = Math.atan2(coord[i+1].y - coord[i].y, coord[i+1].x - coord[i].x);
			Polygon rectEnv = (Polygon)rotation(convexHull, center, -1.0 * angle).getEnvelope();
			
			// Area of the rectangular envelope
			double area = rectEnv.getArea();
			
			// Finding the minimum
			if (area < areaMin) {
				areaMin = area;
				mbr = rectEnv;
				angleForRotation = angle;
			}
		}
		return rotation(mbr, center, angleForRotation);
	}

	
	/** 
	 * (Input parameter: The {@link Polygon} must be a rectangle)
	 * Compute the elongation of a rectangle, equals to the ratio length / width.
	 * If the {@link Polygon} in parameter has a number of coordinates different from 5,
	 * the value -1.0 is returned.
	 * 
	 * @param rectangle a polygon which must be a rectangle
	 * @return the elongation of the rectangle
	 */
	public static double elongation(Polygon rectangle) {
		Coordinate[] coord = rectangle.getCoordinates();
		if (coord.length != 5) return -1.0;
		double lentgh = new LineSegment(coord[0], coord[1]).getLength();
		double lentgh2 = new LineSegment(coord[1], coord[2]).getLength();
		if (lentgh >= lentgh2) return lentgh / lentgh2;
		return lentgh2 / lentgh;
	}
	
	
	/** 
	 * Compute the orientation of a {@link Polygon} (in degrees) comparatively
	 * to the cartographic East (axis X). The orientation is based on the
	 * orientation of the associated minimum bounding rectangle. Its
	 * value is between 0 et 180 degrees, except if the MBR can not
	 * be defined. In that last case, the value -2.0 is returned.
	 * 
	 * @param polygon polygon to process
	 * @return the orientation in degrees, -2.0 if the MBR can not be defined
	 */
	public static double orientation(Polygon polygon) {
		Polygon mbr = PolygonUtil.MBR(polygon);
		if (mbr == null) return -2.0;
		
		Coordinate[] coords = mbr.getCoordinates();
		double length = coords[0].distance(coords[1]);
		double length2 = coords[1].distance(coords[2]);

		Coordinate c1,c2;
		if (length > length2) {
			c1 = coords[0];
			c2 = coords[1];
		} else { // the case length = length2 is here considered
			c1 = coords[1];
			c2 = coords[2];
		}
		double angle = Math.atan2(c1.y - c2.y, c1.x - c2.x);
		
		if (angle < 0) {
			angle += Math.PI;
		}
		
		return (angle * 180.0 / Math.PI);
	}
	
	
	//////////////////////////////////////////////////////////////////
	///////////////////// SHAPE CARACTERISATION //////////////////////
	//////////////////////////////////////////////////////////////////
	
	/** 
	 * Compute if a {@link Polygon} has a circular shape or not, based on
	 * the compactedness index or Miller's index (if Miller's index equals 1,
	 * it means that the shape is a perfect circle).
	 * 
	 * @param polygon polygon to process
	 * @param iMillerThreshold minimum threshold for Miller's index
	 * @return true if the polygon shape is a circle, otherwise false
	 */
	public static boolean isCircle(Polygon polygon, double iMillerThreshold) {
		if (PolygonUtil.compactednessIndex(polygon) >= iMillerThreshold) return true;
		return false;
	}

	
	/** 
	 * Compute if a {@link Polygon} has a rectangular shape or not, based on the comparison
	 * with two shapes: the shape of the minimum bounding rectangle of the {@link Polygon},
	 * and the shape of the convex hull associated to the {@link Polygon}. This comparison
	 * is computed thanks to two thresholds (threshold of maximums), one for each associated
	 * shape. If the shape is rectangular, the orientation (in degrees) based on the is returned; 
	 * otherwise, the value of -1.0 is returned. Finally, if the returned value equals -2.0,
	 * it means that the MBR can not be computed.
	 * 
	 * @param polygon polygon to process
	 * @param sdConvexHullThreshold maximum threshold for surface distance between the polygon and its convex hull
	 * @param sdMBRThreshold maximum threshold for surface distance between the polygon and its associated minimum bounding rectangle
	 * @return the orientation in degrees of the polygon, based on the orientation of the associated minimum bounding rectangle,
	 * -1.0 if the polygon shape is not defined as a retangle, and -2.0 if the MBR can not be computed
	 */
	public static double isRectangle(Polygon polygon, double sdConvexHullThreshold, double sdMBRThreshold) {
		
		Polygon convexHull = (Polygon) polygon.convexHull();
		Polygon mbr = MBR(polygon);
		
		double sdConvexHull = surfaceDistance(polygon, convexHull);
		double sdMBR = surfaceDistance(polygon, mbr);
		
		if (sdConvexHull <= sdConvexHullThreshold && sdMBR <= sdMBRThreshold) {
			return orientation(polygon);
		}
		return -1.0;
	}
	
	
	/** 
	 * Compute if a {@link Polygon} has a rectangular shape or not, based on the comparison
	 * with two shapes: the shape of the minimum bounding rectangle of the {@link Polygon},
	 * and the shape of the convex hull associated to the {@link Polygon}. This comparison
	 * is computed thanks to two thresholds (threshold of maximums), one for each associated
	 * shape.
	 * If the shape is rectangular, the orientation (in degrees) based on the is returned; 
	 * otherwise, the value of -1.0 is returned. Finally, if the returned value equals -2.0,
	 * it means that the MBR can not be computed.
	 * 
 	 * @param sdConvexHull surface distance between the considered polygon and its convex hull
	 * @param sdConvexHullThreshold maximum threshold for surface distance between the considered polygon and its convex hull
	 * @param sdMBR surface distance between the considered polygon and its associated minimum bounding rectangle
	 * @param sdMBRThreshold maximum threshold for surface distance between the considered polygon and its associated minimum bounding rectangle
	 * @param mbr minimum bounding rectangle associated to the considered polygon
	 * @return orientation in degrees of the polygon, based on the orientation of the associated minimum bounding rectangle,
	 * -1.0 if the polygon shape is not defined as a retangle, and -2.0 if the MBR can not be computed
	 */
	public static double isRectangle(double sdConvexHull, double sdConvexHullThreshold,
			double sdMBR, double sdMBRThreshold, Polygon mbr) {
	
		if (sdConvexHull <= sdConvexHullThreshold && sdMBR <= sdMBRThreshold) {
			return orientation(mbr);
		}
		return -1.0;
	}
	
	
	//////////////////////////////////////////////////////////////////
	//////////////////////// POLYGON CLEANING ////////////////////////
	//////////////////////////////////////////////////////////////////
	
	/** 
	 * Clean a {@link Polygon} by removing the unuseful points (in regard to the
	 * chosen threshold) of a {@link Polygon}, i.e. the points which are in the 
	 * alignment between the previous point and the following one. The alignment
	 * is here defined thanks to a given threshold comparatively to Pi (or 180°).
	 * The threshold has to be expressed in degrees.
	 * 
	 * @param polygon polygon to process
	 * @param threshold threshold in degrees to remove points which are aligned
	 * with an angle of 180 +/- threshold in the polygon
	 * @return the new and cleaned polygon
	 */
	public static Polygon cleaningPolygonShape(Polygon polygon, double threshold) {

		// Initialise a Geometryfactory 
		GeometryFactory factory = new GeometryFactory(polygon.getPrecisionModel(), polygon.getSRID());
		
		// Manage the exterior ring
		Coordinate[] coordExt = polygon.getExteriorRing().getCoordinates();
		List<Coordinate> coordsExt = new ArrayList<Coordinate>();
		
		// List of a coordinates list
		for (int i = 0 ; i < coordExt.length ; i++) {
			coordsExt.add(coordExt[i]);
		}
		
		// Dectect unuseful coordinates
		List<Integer> toRemove = new ArrayList<Integer>();
		for (int i = 0 ; i < coordsExt.size() - 2 ; i++) {
			Coordinate c0 = coordExt[i], 
						c1 = coordExt[i + 1], 
						c2 = coordExt[i + 2];
			double angle = Angle.interiorAngle(c0, c1, c2);
			if ( angle >= Math.PI - Math.toRadians(threshold)
					&& angle <= Math.PI + Math.toRadians(threshold)) {
				toRemove.add(0, i + 1);
			}
		}
		
		// Delete unuseful coordinates
		for (Integer i : toRemove) {
			coordsExt.remove(i.intValue());
		}

		// Convert into an array and create the CoordinateSequence
		// associated to the simplified coordinates
		LinearRing lr = factory.createLinearRing(coordsExt.toArray(new Coordinate[coordsExt.size()]));			
		
		// If the polygon has no hole
		if (polygon.getNumInteriorRing() == 0) {
			return new Polygon(lr, null, factory);
		}
		
		// else -- Holes

		LinearRing[] interiorRings = new LinearRing[polygon.getNumInteriorRing()];
		
		// Clean interior rings (holes)
		for (int i = 0 ; i < polygon.getNumInteriorRing() ; i++) {
			Coordinate[] coordInt = polygon.getInteriorRingN(i).getCoordinates();
			List<Coordinate> coordsInt = new ArrayList<Coordinate>();
			
			// List of a coordinates list
			for (int j = 0 ; j < coordInt.length ; j++) {
				coordsInt.add(coordInt[j]);
			}
			
			// Dectect unuseful coordinates
			List<Integer> toRemoveInt = new ArrayList<Integer>();
			for (int k = 0 ; k < coordsInt.size() - 2 ; k++) {
				Coordinate c0 = coordInt[k], 
							c1 = coordInt[k + 1], 
							c2 = coordInt[k + 2];
				double angle = Angle.interiorAngle(c0, c1, c2);
				if ( angle >= Math.PI - Math.toRadians(threshold)
						&& angle <= Math.PI + Math.toRadians(threshold)) {
					toRemoveInt.add(0, k + 1);
				}
			}
			
			// Delete unuseful coordinates
			for (Integer m : toRemoveInt) {
				coordsInt.remove(m.intValue());
			}
			
			// Convert into an array and create the CoordinateSequence
			// associated to the simplified coordinates
			interiorRings[i] = factory.createLinearRing(coordsExt.toArray(new Coordinate[coordsExt.size()]));
		}
		return new Polygon(lr, interiorRings, factory);			

	}
	
	
	/** 
	 * Clean a {@link MultiPolygon} by removing the unuseful points (in regard to the
	 * chosen threshold) of a {@link MultiPolygon}, i.e. the points which are in the 
	 * alignment between the previous point and the following one. The alignment
	 * is here defined thanks to a given threshold comparatively to Pi (or 180°).
	 * The threshold has to be expressed in degrees.
	 * 
	 * @param multiPolygon multipolygon to process
	 * @param threshold threshold in degrees to remove points which are aligned
	 * with an angle of 180 +/- threshold in the multipolygon
	 * @return the new and cleaned multipolygon
	 */
	public static MultiPolygon cleaningMultiPolygonShape(MultiPolygon multiPolygon, double threshold) {
		
		Polygon[] polygons = new Polygon[multiPolygon.getNumGeometries()];
		
		// Use the cleaning method cleaningPolygonShape
		for (int i = 0 ; i < multiPolygon.getNumGeometries() ; i++) {
			Polygon polygon = (Polygon)multiPolygon.getGeometryN(i);
			polygons[i] = cleaningPolygonShape(polygon, threshold);
		}
		
		// Initialise a Geometryfactory and return a new and clean multipolygon
		return new MultiPolygon(polygons, new GeometryFactory(multiPolygon.getPrecisionModel(), multiPolygon.getSRID()));
	}

	
	//////////////////////////////////////////////////////////////////
	//////////////////////  POLYGON CLEANING   ///////////////////////
	////////////////////// AUTOMATIC THRESHOLD ///////////////////////
	//////////////////////////////////////////////////////////////////
	
	/** 
	 * Clean a {@link Polygon} by removing the unuseful points (in regard to the
	 * chosen threshold) of a {@link Polygon}, i.e. the points which are in the 
	 * alignment between the previous point and the following one. The alignment
	 * is here defined thanks to a threshold comparatively to Pi (or 180°)
	 * which is automatically computed (quasi optimal, considering a maximum
	 * area variation of 1%, comparating area variations, not the global one).
	 * The threshold has a maximum of 2 degrees.
	 * 
	 * TODO: Check if the threshold matches a reality
	 * 
	 * @param polygon polygon to clean
	 * @return the new and cleaned polygon
	 */
	public static Polygon cleaningPolygonShapeAutomaticThreshold(Polygon polygon) {

		// Initialise the new polygon
		Polygon newPolygon = null;
		
		// 1% of the original area of the polygon
		double areaStart = 0.01 * polygon.getArea();
		double areaCurrent = 0;
		
		// Threshold with a maximum of 2 degrees
		double thresholdMax = Math.toRadians(2);
		double thresholdMin = 0;
		
		double thresholdTest;
		
		boolean flag = true;
		boolean flagCondition;

		// Dichotomy to find the most adapted threshold
		while(flag) {
			thresholdTest = (thresholdMax + thresholdMin) / 2.0;
			newPolygon = cleaningPolygonShape(polygon, thresholdTest);
			areaCurrent = newPolygon.symDifference(polygon).getArea();
			if (areaCurrent <= areaStart) {
				thresholdMin = thresholdTest;
				flagCondition = true;
			}
			else {
				thresholdMax = thresholdTest;
				flagCondition = false;
			}
			// Condition to exit from the dichotomous process: 
			// the difference between the two angular thresholds has
			// to be less than 0.1 degree
			if (thresholdMax - thresholdMin <= Math.toRadians(0.1) && flagCondition) {
				flag = false;
			}
		}
		return newPolygon;
	}
	
	
	/** 
	 * Clean a {@link MultiPolygon} by removing the unuseful points (in regard to the
	 * chosen threshold) of a {@link MultiPolygon}, i.e. the points which are in the 
	 * alignment between the previous point and the following one. The alignment
	 * is here defined thanks to a threshold comparatively to Pi (or 180°)
	 * which is automatically computed (quasi optimal, considering a maximum
	 * area variation of 1%, comparating area variations, not the global one).
	 * The threshold has a maximum of 2 degrees.
	 * 
	 * TODO: Check if the threshold matches a reality
	 * 
	 * @param multiPolygon multipolygon to clean
	 * @return the new and cleaned multipolygon
	 */
	public static MultiPolygon cleaningMultiPolygonShapeAutomaticThreshold(
			MultiPolygon multiPolygon) {

		Polygon[] polygons = new Polygon[multiPolygon.getNumGeometries()];
		
		// Use the cleaning method cleaningPolygonShapeAutomaticThreshold
		for (int i = 0 ; i < multiPolygon.getNumGeometries() ; i++) {
			Polygon polygon = (Polygon)multiPolygon.getGeometryN(i);
			polygons[i] = cleaningPolygonShapeAutomaticThreshold(polygon);
		}
		
		// Initialise a Geometryfactory and return a new and clean multipolygon
		return new MultiPolygon(polygons, new GeometryFactory(multiPolygon.getPrecisionModel(), multiPolygon.getSRID()));
	}
		
}