/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.plugin.geometry;

import java.util.Iterator;

import javax.swing.JOptionPane;

import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.task.TaskMonitor;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.StandardCategoryNames;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.plugin.ThreadedPlugIn;
import com.vividsolutions.jump.workbench.ui.ErrorDialog;
import com.vividsolutions.jump.workbench.ui.GUIUtil;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;
import com.vividsolutions.jump.workbench.ui.plugin.FeatureInstaller;

import fr.ign.cogit.alpage.morphal.i18n.I18NPlug;

/** 
 * Plugin to create a {@link Layer} of valid geometric {@link Feature}s (and also another
 * with non valid geometric {@link Feature}s) from an original {@link Layer} which can
 * contain {@link Feature}s with any geometric types
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 *
 * @version 0.1 (2011-01-09)
 */
public class GeometryValidatorPlugin extends AbstractPlugIn implements ThreadedPlugIn {

	////////////////////////////////////////////////////////////////////////////////////
	// Internationalisation
	
	private final static String MENU = 	I18NPlug.getI18N("project");
	private final static String MENU_SUBTITLE = I18NPlug.getI18N("geometry");
	private final static String TITLE = I18NPlug.getI18N("geometry.ValidatorPlugin.title");
	private final static String DESCRIPTION = I18NPlug.getI18N("geometry.ValidatorPlugin.description");

	private final static String LAYER = I18NPlug.getI18N("geometry.ValidatorPlugin.layer");

	private final static String VALID_ALL = I18NPlug.getI18N("geometry.ValidatorPlugin.allValid");
	
	private final static String WARNING_START = I18NPlug.getI18N("geometry.ValidatorPlugin.warningStart");
	private final static String WARNING_CONTAINS = I18NPlug.getI18N("geometry.ValidatorPlugin.warningContains");
	private final static String WARNING_VALID = I18NPlug.getI18N("geometry.ValidatorPlugin.warningValidFeatures");
	private final static String WARNING_LAYER = I18NPlug.getI18N("geometry.ValidatorPlugin.warningLayer");
	private final static String WARNING_NON_VALID = I18NPlug.getI18N("geometry.ValidatorPlugin.warningNonValid");
	
	private final static String ERROR = I18NPlug.getI18N("error");
	private final static String EMPTY = I18NPlug.getI18N("error.emptyCollection");	

	////////////////////////////////////////////////////////////////////////////////////
	
	private static MultiInputDialog dialog;
	
	/** 
	 * Initialisation.
	 * 
	 * @param context
	 */
	@Override
	public void initialize(PlugInContext context) {
		FeatureInstaller featureInstaller = new FeatureInstaller(context.getWorkbenchContext());
		featureInstaller.addMainMenuItem(
				this,
				new String[] { MENU, MENU_SUBTITLE },
				TITLE,
				false,
				null,
				new MultiEnableCheck()
				.add(context.getCheckFactory().createTaskWindowMustBeActiveCheck())
				.add(context.getCheckFactory().createAtLeastNLayersMustExistCheck(1)));
	}


	/** 
	 * Global execution.
	 * 
	 * @param context
	 * @return true if process is executed, otherwise false
	 */
	@Override
	public boolean execute(PlugInContext context) {
		this.reportNothingToUndoYet(context);
		dialog = new MultiInputDialog(
				context.getWorkbenchFrame(), getName(), true);
		setDialogValues(dialog, context);
		GUIUtil.centreOnWindow(dialog);
		dialog.setVisible(true);
		if (! dialog.wasOKPressed()) { return false; }
		return true;
	}

	
	/** 
	 * Initialisation of the dialog interface.
	 * 
	 * @param dialog
	 * @param context
	 */
	private void setDialogValues(MultiInputDialog dialog, PlugInContext context) {

		dialog.setSideBarDescription(DESCRIPTION);

		dialog.addLayerComboBox(LAYER,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());
	}


	/** 
	 * Execution.
	 * 
	 * @param monitor
	 * @param context
	 */
	public void run(TaskMonitor monitor, PlugInContext context) {
		validation(context);
		System.gc();
	}


	/** 
	 * Validation.
	 * 
	 * @param context
	 * @return true if the process can be launched, otherwise false
	 */
	private boolean validation(PlugInContext context) {
		
		// Layer to process
		Layer layer = dialog.getLayer(LAYER);

		// Features of the layer to process
		final FeatureCollection collection = layer.getFeatureCollectionWrapper();

		// If the collection is empty
		if (collection.isEmpty()) {
			ErrorDialog.show(context.getWorkbenchFrame(),
					ERROR,
					EMPTY,
					"");
			return false;
		}

		validateCollection(context, collection, layer.getName());
		return true;
	}

	
	/** 
	 * Create and display two {@link Layer}s from the splitting of a collection:
	 * a layer with valid geometric {@link Feature}s, and another with non valid
	 * geometric {@link Feature}s.
	 * If all {@link Feature}s of the collection are valid, display a message
	 * to inform the user about it, and do not display any valid or non valid
	 * {@link Layer}.
	 * 
	 * @param context
	 * @param collection collection to validate geometrically
	 * @return true
	 */
	@SuppressWarnings({ "unchecked" })
	private boolean validateCollection(PlugInContext context, FeatureCollection collection, String layerName) {

		FeatureSchema fsValid = (FeatureSchema) collection.getFeatureSchema().clone();
		FeatureSchema fsNonValid = (FeatureSchema) collection.getFeatureSchema().clone();

		FeatureDataset collectionValid = new FeatureDataset(fsValid);		
		FeatureDataset collectionNonValid = new FeatureDataset(fsNonValid);		
		
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature feature = it.next();
			
			if (feature.getGeometry().isValid()) {
				Feature fValide = new BasicFeature(fsValid);
				for (int i = 0 ; i < feature.getAttributes().length ; i++) {
					fValide.setAttribute(i, feature.getAttribute(i));
				}
				collectionValid.add(fValide);
			} else {
				Feature fNonValide = new BasicFeature(fsNonValid);
				for (int i = 0 ; i < feature.getAttributes().length ; i++) {
					fNonValide.setAttribute(i, feature.getAttribute(i));
				}
				collectionNonValid.add(fNonValide);		
			}
		}

		// All features are valid
		if (collectionValid.size() == collection.size()) {
			JOptionPane.showMessageDialog(null, VALID_ALL, "Validation", JOptionPane.INFORMATION_MESSAGE);
		} else { // Some features are invalid => create layers
			context.addLayer(StandardCategoryNames.RESULT, layerName + "_valid", collectionValid);
			context.addLayer(StandardCategoryNames.RESULT, layerName + "_non_valid", collectionNonValid);
			String message = WARNING_START + " "
				+ layerName + "_valid " + WARNING_CONTAINS + " " + collectionValid.size() + " " + WARNING_VALID
				+ WARNING_LAYER + " " + layerName + "_non_valid " + WARNING_CONTAINS + " " + collectionNonValid.size() + " " + WARNING_NON_VALID;
			JOptionPane.showMessageDialog(null, message, "Validation", JOptionPane.WARNING_MESSAGE);
		}

		return true;
	}
	
}