/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.plugin.linestring;

import java.util.Iterator;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.task.TaskMonitor;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.StandardCategoryNames;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.plugin.ThreadedPlugIn;
import com.vividsolutions.jump.workbench.ui.ErrorDialog;
import com.vividsolutions.jump.workbench.ui.GUIUtil;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;

import fr.ign.cogit.alpage.morphal.geometry.LineStringUtil;
import fr.ign.cogit.alpage.morphal.i18n.I18NPlug;

/** 
 * Plugin to clean a {@link Layer} which contains features with a {@link LineString}
 * or/and a {@link MultiLineString} geometric type.
 * 
 * Attention: the {@link Feature}s with a {@link Geometry} different from a {@link LineString}
 * or a {@link MultiLineString} are not computed.
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 * 
 * @version 0.1 (2011-01-09)
 */
public class LineStringCleaningPlugin extends AbstractPlugIn implements ThreadedPlugIn {
	
	private static Logger LOGGER = Logger.getLogger(LineStringCleaningPlugin.class);

	////////////////////////////////////////////////////////////////////////////////////
	// Internationalisation
	
	private final static String MENU = 	I18NPlug.getI18N("project");
	private final static String MENU_SUBTITLE = I18NPlug.getI18N("linestring");
	private final static String TITLE = I18NPlug.getI18N("linestring.LineStringCleaningPlugin.title");
	private final static String DESCRIPTION = I18NPlug.getI18N("linestring.LineStringCleaningPlugin.description");
	
	private final static String LAYER = I18NPlug.getI18N("linestring.LineStringCleaningPlugin.layer");
	private final static String THRESHOLD_ANGLE = I18NPlug.getI18N("linestring.LineStringCleaningPlugin.thresholdAngle");
	
	private final static String ERROR = I18NPlug.getI18N("error");
	private final static String EMPTY = I18NPlug.getI18N("error.emptyCollection");
	private final static String ERROR_INCORRECT_TYPE_TITLE = I18NPlug.getI18N("error.linestring.incorrectGeometricTypeTitle");
	private final static String ERROR_INCORRECT_TYPE = I18NPlug.getI18N("error.linestring.incorrectGeometricType");
	private final static String ERROR_INCORRECT_TYPE_ALL = I18NPlug.getI18N("error.linestring.incorrectGeometricTypeAll");

	////////////////////////////////////////////////////////////////////////////////////
	
	private static MultiInputDialog dialog;	

	
	/** 
	 * Initialisation.
	 * 
	 * @param context
	 */
	@Override
	public void initialize(PlugInContext context) {
		context.getFeatureInstaller().addMainMenuItem(
				this,
				new String[] { MENU, MENU_SUBTITLE },
				TITLE,
				false,
				null,
				new MultiEnableCheck()
				.add(context.getCheckFactory().createTaskWindowMustBeActiveCheck())
				.add(context.getCheckFactory().createAtLeastNLayersMustExistCheck(1)));
	}

	
	/** 
	 * Global execution.
	 * 
	 * @param context
	 * @return true if the process is executed, otherwise false
	 */
	@Override
	public boolean execute(PlugInContext context) {
		this.reportNothingToUndoYet(context);
		dialog = new MultiInputDialog(context.getWorkbenchFrame(), getName(), true);
		setDialogValues(dialog, context);
		GUIUtil.centreOnWindow(dialog);
		dialog.setVisible(true);
		if (! dialog.wasOKPressed()) { return false; }
		return true;
	}

	
	/** 
	 * Initialisation of the dialog interface.
	 * 
	 * @param dialog
	 * @param context
	 */
	private void setDialogValues(MultiInputDialog dialog, PlugInContext context) {
		dialog.setSideBarDescription(DESCRIPTION);

		dialog.addLayerComboBox(LAYER,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());
		
		dialog.addDoubleField(THRESHOLD_ANGLE, 1, 4, THRESHOLD_ANGLE);
	}

	
	/** 
	 * Execution.
	 * 
	 * @param monitor
	 * @param context
	 */
	public void run(TaskMonitor monitor, PlugInContext context) {
		cleanLineString(context);
		System.gc();
	}

	
	/** 
	 * Clean a layer of {@link LineString] or {@link MultiLineString}.
	 * 
	 * @param context
	 * @return true if the process can be launched, otherwise false
	 */
	@SuppressWarnings("unchecked")
	private boolean cleanLineString(PlugInContext context) {
		
		// Layer to process
		Layer layer = dialog.getLayer(LAYER);

		// Features of the layer to process
		final FeatureCollection collection = layer.getFeatureCollectionWrapper();

		if (collection.isEmpty()) {
			ErrorDialog.show(
					context.getWorkbenchFrame(),
					ERROR,
					EMPTY,
					"");
			return false;
		}
		
		// Management of incorrect geometric input (other than LineString or MultiLineString)
		// This preprocess is not long and allows to users to make a choice about what they
		// want to do.
		int nLML = 0; // Number of features with a LineString or a MultiLineString geometric type
		
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature f = it.next();
			if (f.getGeometry() instanceof LineString || 
					f.getGeometry() instanceof MultiLineString) {
				nLML++;
			}
		}
		
		int size = collection.size();
		
		if (nLML == size) { // All features have a correct type
			cleanLineStringCollection(context, collection);
		} else if (nLML == 0) { // All features have an incorrect type
			ErrorDialog.show(
					context.getWorkbenchFrame(),
					ERROR_INCORRECT_TYPE_TITLE,
					ERROR_INCORRECT_TYPE_ALL,
					"");
			return false;
		} else {
			int result = JOptionPane.showConfirmDialog(
					context.getWorkbenchFrame(),
					(size - nLML) + " " + ERROR_INCORRECT_TYPE,
					ERROR_INCORRECT_TYPE_TITLE,
					JOptionPane.OK_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE);
			
			// If OK, the cleaning process is launched. Otherwise, it is not.
			if (result == JOptionPane.OK_OPTION) {
				cleanLineStringCollection(context, collection);
			} else {
				return false;
			}
		}
		return true;
	}
	
	
	/** 
	 * Clean a layer of {@link LineString] or {@link MultiLineString}.
	 * 
	 * @param context
	 * @param collection
	 * @return true
	 */
	@SuppressWarnings("unchecked")
	private boolean cleanLineStringCollection(PlugInContext context, FeatureCollection collection) {

		// Layer to process
		Layer layer = dialog.getLayer(LAYER);
		
		// Threshold
		double threshold = dialog.getDouble(THRESHOLD_ANGLE);
		if (threshold < 0) {
			threshold = -threshold;
		}
		
		// New cleaned layer
		FeatureSchema fsClone = (FeatureSchema) collection.getFeatureSchema().clone();
		FeatureDataset collectionNew = new FeatureDataset(fsClone);

		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			
			Feature feature = it.next();

			Feature featureNew = new BasicFeature(fsClone);
			for (int i = 0 ; i < feature.getAttributes().length ; i++) {
				featureNew.setAttribute(i, feature.getAttribute(i));
			}
			
			if (featureNew.getGeometry() instanceof LineString) {
				LineString line = LineStringUtil.cleaningLineShape((LineString)feature.getGeometry(), threshold);
				featureNew.setGeometry(line);
				collectionNew.add(featureNew);
			} else if (featureNew.getGeometry() instanceof MultiLineString) {
				MultiLineString multiLine = LineStringUtil.cleaningMultiLineShape((MultiLineString)feature.getGeometry(), threshold);
				featureNew.setGeometry(multiLine);
				collectionNew.add(featureNew);
			} else {
				LOGGER.warn("Feature ID = " + feature.getID() + " is not process: not a LineString or a MultiLineString.");
			}			
		}		

		// The new collection should normally not be empty -- due to the checking in the initial part (method cleanLineString) --
		if (!collectionNew.isEmpty()) {
			context.addLayer(StandardCategoryNames.RESULT, layer.getName() + "_simplify", collectionNew);
		}
		
		return true;
	}

}