/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.plugin.linestring;

import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.CoordinateSequence;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineSegment;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.impl.CoordinateArraySequenceFactory;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.task.TaskMonitor;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.StandardCategoryNames;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.plugin.ThreadedPlugIn;
import com.vividsolutions.jump.workbench.ui.ErrorDialog;
import com.vividsolutions.jump.workbench.ui.GUIUtil;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;
import com.vividsolutions.jump.workbench.ui.plugin.FeatureInstaller;

import fr.ign.cogit.alpage.morphal.i18n.I18NPlug;

/** 
 * Plugin to create a {@link Layer} of segments from a {@link Layer} of {@link LineString}
 * or {@link MultiLineString} {@link Feature}s.
 *
 * Attention: the {@link Feature}s with a {@link Geometry} different from
 * a {@link LineString} or a {@link MultiLineString} are not computed.
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 *
 * @version 0.1 (2011-01-09)
 */
public class LineStringToSegmentsPlugin extends AbstractPlugIn implements ThreadedPlugIn {
	
	private static Logger LOGGER = Logger.getLogger(LineStringToSegmentsPlugin.class);
	
	////////////////////////////////////////////////////////////////////////////////////
	// Internationalisation
	
	private final static String MENU = I18NPlug.getI18N("project");
	private final static String MENU_SUBTITLE = I18NPlug.getI18N("linestring");
	private final static String TITLE = I18NPlug.getI18N("linestring.LineStringToSegmentsPlugin.title");
	private final static String DESCRIPTION = I18NPlug.getI18N("linestring.LineStringToSegmentsPlugin.description");
	
	private final static String LAYER = I18NPlug.getI18N("linestring.LineStringToSegmentsPlugin.layer");
	
	private final static String UNICITY = I18NPlug.getI18N("linestring.LineStringToSegmentsPlugin.unicity");

	private final static String ERROR = I18NPlug.getI18N("error");
	private final static String EMPTY = I18NPlug.getI18N("error.emptyCollection");
	private final static String ERROR_INCORRECT_TYPE_TITLE = I18NPlug.getI18N("error.linestring.incorrectGeometricTypeTitle");
	private final static String ERROR_INCORRECT_TYPE = I18NPlug.getI18N("error.linestring.incorrectGeometricType");
	private final static String ERROR_INCORRECT_TYPE_ALL = I18NPlug.getI18N("error.linestring.incorrectGeometricTypeAll");

	////////////////////////////////////////////////////////////////////////////////////
	
	private static MultiInputDialog dialog;
	private Map<LineSegment, Integer> segmentMap = new TreeMap<LineSegment, Integer>();
	private boolean unicity = true;


	/** 
	 * Initialisation.
	 * 
	 * @param context
	 */
	@Override
	public void initialize(PlugInContext context) {
		FeatureInstaller featureInstaller = new FeatureInstaller(context.getWorkbenchContext());
		featureInstaller.addMainMenuItem(
				this,
				new String[] { MENU, MENU_SUBTITLE },
				TITLE,
				false,
				null,
				new MultiEnableCheck()
				.add(context.getCheckFactory().createTaskWindowMustBeActiveCheck())
				.add(context.getCheckFactory().createAtLeastNLayersMustExistCheck(1)));
	}
	
	
	/** 
	 * Global execution.
	 * 
	 * @param context
	 * @return true if the process is executed, otherwise false
	 */
	@Override
	public boolean execute(PlugInContext context) {
		this.reportNothingToUndoYet(context);
		dialog = new MultiInputDialog(context.getWorkbenchFrame(), getName(), true);
		setDialogValues(dialog, context);
		GUIUtil.centreOnWindow(dialog);
		dialog.setVisible(true);
		if (! dialog.wasOKPressed()) { return false; }
		return true;
	}

	
	/** 
	 * Initialisation of the dialog interface.
	 * 
	 * @param dialog
	 * @param context
	 */
	private void setDialogValues(MultiInputDialog dialog, PlugInContext context) {
		dialog.setSideBarDescription(DESCRIPTION);

		dialog.addLayerComboBox(LAYER,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());

		// Choice about unicity of computed segments
		dialog.addCheckBox(UNICITY, unicity);
	}

	
	/** 
	 * Execution.
	 * 
	 * @param monitor
	 * @param context
	 */
	public void run(TaskMonitor monitor, PlugInContext context) {
		unicity = dialog.getBoolean(UNICITY);
		linestringToSegmentsGlobal(context);
		System.gc();
	}

	
	/** 
	 * Initialisation of the process, including the management of features constraints
	 * to take into account.
	 * 
	 * @param context
	 * @return true if the process is executed, otherwise false
	 */
	@SuppressWarnings("unchecked")
	private boolean linestringToSegmentsGlobal(PlugInContext context) {

		// Layer to process
		Layer layer = dialog.getLayer(LAYER);

		// Features of the layer to process
		final FeatureCollection collection = layer.getFeatureCollectionWrapper();

		// If the collection is empty
		if (collection.isEmpty()) {
			ErrorDialog.show(context.getWorkbenchFrame(),
					ERROR,
					EMPTY,
					"");
			return false; 
		}

		// Management of incorrect geometric input (other than LineString or MultiLineString).
		// This preprocess is not long and allows to users to make a choice about what they
		// want to do
		int nLML = 0; // Number of features with a LineString or a MultiLineString geometric type
		
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature f = it.next();
			if (f.getGeometry() instanceof LineString || 
					f.getGeometry() instanceof MultiLineString) {
				nLML++;
			}
		}
		
		int size = collection.size();
		
		if (nLML == size) { // All features have a correct type
			lineStringToSegments(context, layer, collection);
		} else if (nLML == 0) { // All features have an incorrect type
			ErrorDialog.show(
					context.getWorkbenchFrame(),
					ERROR_INCORRECT_TYPE_TITLE,
					ERROR_INCORRECT_TYPE_ALL,
					"");
			return false;
		} else {
			int result = JOptionPane.showConfirmDialog(
					context.getWorkbenchFrame(),
					(size - nLML) + " " + ERROR_INCORRECT_TYPE,
					ERROR_INCORRECT_TYPE_TITLE,
					JOptionPane.OK_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE);
			
			// If OK, the splitting process is launched. Otherwise, it is not.
			if (result == JOptionPane.OK_OPTION) {
				lineStringToSegments(context, layer, collection);
			} else {
				return false;
			}
		}
		return true;
	}

	
	/** 
	 * Create a new {@link Layer} with segments resulting from the split of a collection
	 * of {@link LineString} and/or {@link MultiLineString}.
	 * 
	 * @param context
	 * @param collection
	 * @return true
	 */
	@SuppressWarnings("unchecked")
	private boolean lineStringToSegments(PlugInContext context, Layer layer, FeatureCollection collection) {

		FeatureSchema fsClone = (FeatureSchema) collection.getFeatureSchema().clone();
		FeatureDataset collectionSegments = new FeatureDataset(fsClone);
				
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			
			Feature feature = it.next();
			
			if (feature.getGeometry() instanceof LineString) {
				LineString geom = (LineString) feature.getGeometry();
				lineStringToCollection(collectionSegments, feature, geom);
			} else if (feature.getGeometry() instanceof MultiLineString) {
				MultiLineString geom = (MultiLineString) feature.getGeometry();
				
				for (int i = 0 ; i < geom.getNumGeometries() ; i++) {
					lineStringToCollection(collectionSegments, feature, (LineString)geom.getGeometryN(i));
				}
			} else {
				LOGGER.warn("Feature ID " + feature.getID() + " is not process: not a LineString or a MultiLineString.");
			}		
		}

		context.addLayer(StandardCategoryNames.RESULT, layer.getName() + "_segments", collectionSegments);

		return true;
	}

	/**
	 * Split a {@link LineString} into segments.
	 * 
 	 * @param dataSet dataset to be able to add the created {@link Feature}s
	 * @param feature {@link Feature} to split
	 * @param geom the {@link Geometry} of the {@link Feature} if it is a {@link LineString}
	 * or the i-th {@link Geometry} of the {@link Feature} if it is originally a {@link MultiLineString}
	 */
	private void lineStringToCollection(FeatureDataset dataSet, Feature feature, LineString geom) {
		GeometryFactory gf = new GeometryFactory(geom.getPrecisionModel(), geom.getSRID()); 

		Coordinate[] coord = geom.getCoordinates();
		for (int j = 0 ; j < coord.length - 1; j++) {
			addFeature(dataSet, feature, gf,coord[j], coord[j + 1]);
		}
	}

	
	/** 
	 * Add a feature which is a segment (defined by the two {@link Coordinate} in input)
	 * to the segments list. Manage the unicity parameter given by the user.
	 * 
	 * @param dataSet dataset to be able to add the created {@link Feature}
	 * @param feature {@link Feature} to add
	 * @param gf geometry factory to be able to create a {@link LineString} in link with
	 * the {@link Coordinate} in input
	 * @param coord first {@link Coordinate} of the segment
	 * @param coord2 second {@link Coordinate} of the segment
	 */
	private void addFeature(FeatureDataset dataSet, Feature feature, GeometryFactory gf, 
			Coordinate coord, Coordinate coord2) {
		
		Coordinate[] c = {coord, coord2};
		CoordinateSequence seq = CoordinateArraySequenceFactory.instance().create(c);

		// If unicity required
		if (unicity) {
			LineSegment segment = new LineSegment(c[0], c[1]);
			segment.normalize();

			if (segmentMap.get(segment) == null) {
				segmentMap.put(segment, 1); 

				Feature featureNew = new BasicFeature(dataSet.getFeatureSchema());

				// Copy the attributes (except the geometry)
				for(int k = 1 ; k < featureNew.getAttributes().length ; k++) {
					featureNew.setAttribute(k, feature.getAttribute(k));
				}

				featureNew.setGeometry(new LineString(seq,gf));
				dataSet.add(featureNew);
			}			
		} else { // If non unicity
			Feature featureNew = new BasicFeature(dataSet.getFeatureSchema());

			// Copy the attributes (except the geometry)
			for(int k = 1 ; k < featureNew.getAttributes().length ; k++) {
				featureNew.setAttribute(k, feature.getAttribute(k));
			}

			featureNew.setGeometry(new LineString(seq,gf));
			dataSet.add(featureNew);	
		}		
	}

}