/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.plugin.linestring;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JOptionPane;
import javax.swing.JTextField;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.apache.log4j.Logger;
import org.jfree.ui.RefineryUtilities;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jump.feature.AttributeType;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.task.TaskMonitor;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.LayerEventType;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.plugin.ThreadedPlugIn;
import com.vividsolutions.jump.workbench.ui.ErrorDialog;
import com.vividsolutions.jump.workbench.ui.GUIUtil;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;
import com.vividsolutions.jump.workbench.ui.plugin.FeatureInstaller;

import fr.ign.cogit.alpage.morphal.geometry.AngleUtil;
import fr.ign.cogit.alpage.morphal.geometry.GeometryUtil;
import fr.ign.cogit.alpage.morphal.i18n.I18NPlug;
import fr.ign.cogit.alpage.morphal.ui.statistics.Histogram;

/** 
 * Plugin to compute and add three new attributes to the specified {@link Layer} of segments:
 * 1. length (attribute: "length"): length of the segment,
 * 2. orientation (attribute: "orientatio"): orientation of the segment,
 * 3. classification (attribute: "classe") -- optional --: classification of the computed orientations.
 * 
 * Attention: the {@link Feature}s with a {@link Geometry} different from a segment
 * are not computed.
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 * 
 * @version 0.1 (2011-01-09)
 */
public class SegmentOrientationPlugin extends AbstractPlugIn implements ThreadedPlugIn {
	
	private static Logger LOGGER = Logger.getLogger(SegmentOrientationPlugin.class);
	
	////////////////////////////////////////////////////////////////////////////////////
	// Internationalisation
	
	private final static String MENU = 	I18NPlug.getI18N("project");
	private final static String MENU_SUBTITLE = I18NPlug.getI18N("linestring");
	private final static String TITLE = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.title");
	private final static String DESCRIPTION = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.description");

	private final static String LAYER = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.layer");

	private final static String ACCURACY = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.accuracy");
	
	private final static String UNIT = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.unit");
	private final static String U_DEGREE = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.degree");
	private final static String U_RADIAN = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.radian");
	private final static String U_GRADE = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.grade");
	private final static String INTERVAL = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.interval");
	
	private final static String HISTOGRAM = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.histogram");
	private final static String HISTOGRAM_STEP = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.histogramStep");
	private final static String TITLE_HISTOGRAM = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.histogramTitle");
	
	private final static String CLASSIFICATION = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.classification");
	private final static String CLASSIFICATION_STEP = I18NPlug.getI18N("linestring.SegmentOrientationPlugin.classificationStep");

	private final static String ERROR = I18NPlug.getI18N("error");
	private final static String EMPTY = I18NPlug.getI18N("error.emptyCollection");
	private final static String ERROR_INCORRECT_TYPE_TITLE = I18NPlug.getI18N("error.segment.incorrectGeometricTypeTitle");
	private final static String ERROR_INCORRECT_TYPE = I18NPlug.getI18N("error.segment.incorrectGeometricType");
	private final static String ERROR_INCORRECT_TYPE_ALL = I18NPlug.getI18N("error.segment.incorrectGeometricTypeAll");

	////////////////////////////////////////////////////////////////////////////////////
	
	private static MultiInputDialog dialog;

	private JComboBox jComboInterval, jComboUnits;
	
	private JCheckBox jcbHistogram, jcbClassification;
	private JTextField histogramStep, classificationStep;
	
	private double[] orientations;


	/** 
	 * Initialisation.
	 * 
	 * @param context
	 */
	@Override
	public void initialize(PlugInContext context) {
		FeatureInstaller featureInstaller = new FeatureInstaller(context.getWorkbenchContext());
		featureInstaller.addMainMenuItem(
				this,
				new String[] { MENU, MENU_SUBTITLE },
				TITLE,
				false,
				null,
				new MultiEnableCheck()
				.add(context.getCheckFactory().createTaskWindowMustBeActiveCheck())
				.add(context.getCheckFactory().createAtLeastNLayersMustExistCheck(1)));
	}
	
	
	/** 
	 * Global execution.
	 * 
	 * @param context
	 * @return true if process is executed, otherwise false
	 */
	@Override
	public boolean execute(PlugInContext context) {
		this.reportNothingToUndoYet(context);
		dialog = new MultiInputDialog(context.getWorkbenchFrame(), getName(), true);
		setDialogValues(dialog, context);
		GUIUtil.centreOnWindow(dialog);
		dialog.setVisible(true);
		if (! dialog.wasOKPressed()) { return false; }		
		return true;
	}

	
	/** 
	 * Initialisation of the dialog interface.
	 * 
	 * @param dialog
	 * @param context
	 */
	private void setDialogValues(MultiInputDialog dialog, PlugInContext context) {
		dialog.setSideBarDescription(DESCRIPTION);

		dialog.addLayerComboBox(LAYER,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());

		// Choice of unit
		Collection<String> items = new ArrayList<String>();		
		items.add(U_DEGREE);
		items.add(U_RADIAN);
		items.add(U_GRADE);
		jComboUnits = dialog.addComboBox(UNIT, U_DEGREE, items, "");

		// Choice of an interval for the computation of angles
		Collection<String> itemsPi = new ArrayList<String>();
		itemsPi.add("[0 ; Pi[");
		itemsPi.add("[0 ; Pi/2[");
		jComboInterval = dialog.addComboBox(INTERVAL, "[0 ; Pi[", itemsPi, "");
		
		dialog.addCheckBox(ACCURACY, false);
		
		// Histogram
		this.jcbHistogram = dialog.addCheckBox(HISTOGRAM, false);
		this.jcbHistogram.addChangeListener(new HistogramChange());
		this.histogramStep = dialog.addIntegerField(HISTOGRAM_STEP, 5, 4, HISTOGRAM_STEP);
		this.histogramStep.setEnabled(false);
		
		// Classification
		this.jcbClassification = dialog.addCheckBox(CLASSIFICATION, false);
		this.jcbClassification.addChangeListener(new ClassificationChange());
		this.classificationStep = dialog.addDoubleField(CLASSIFICATION_STEP, 10.0, 5, CLASSIFICATION_STEP);
		this.classificationStep.setEnabled(false);
	}
	
	
	/** Internal class: management of the histogram step checkbox. */	
	class HistogramChange implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			boolean flag = dialog.getCheckBox(HISTOGRAM).isSelected();
			histogramStep.setEnabled(flag);
		}
	}
	
	
	/** Internal class: management of the classification step checkbox. */	
	class ClassificationChange implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			boolean flag = dialog.getCheckBox(CLASSIFICATION).isSelected();
			classificationStep.setEnabled(flag);
		}
	}
	

	/** 
	 * Execution.
	 * 
	 * @param monitor
	 * @param context
	 */
	public void run(TaskMonitor monitor, PlugInContext context) {
		orientation(context);
		System.gc();
	}

	
	/** 
	 * Initialisation of the orientation computation.
	 *  
	 * @param context
	 * @return true if the process can be launched, otherwise false
	 */
	@SuppressWarnings("unchecked")
	private boolean orientation(PlugInContext context) {
		
		// Layer to process
		Layer layer = dialog.getLayer(LAYER);

		// Features of the layer to process
		final FeatureCollection collection = layer.getFeatureCollectionWrapper();

		// If the collection is empty
		if (collection.isEmpty()) {
			ErrorDialog.show(context.getWorkbenchFrame(),
					ERROR,
					EMPTY,
					"");
			return false;
		}

		// Management of incorrect geometric input (other than a segment -- two coordinates --)
		// This preprocess is not long and allows to users to make a choice about what they
		// want to do.
		int nS = 0; // Number of features with a "segment" geometric type
		
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature f = it.next();
			Geometry geom = f.getGeometry();
			if ((geom instanceof LineString || (geom instanceof MultiLineString && geom.getNumGeometries() ==1))
					&& geom.getCoordinates().length == 2) {
				nS++;
			}
		}
		
		int size = collection.size();
		
		if (nS == size) { // All features have a correct type
			orientations = new double[collection.size()];
			orientationSegments(layer, collection);
		} else if (nS == 0) { // All features have an incorrect type
			ErrorDialog.show(
					context.getWorkbenchFrame(),
					ERROR_INCORRECT_TYPE_TITLE,
					ERROR_INCORRECT_TYPE_ALL,
					"");
			return false;
		} else {
			int result = JOptionPane.showConfirmDialog(
					context.getWorkbenchFrame(),
					(size - nS) + " " + ERROR_INCORRECT_TYPE,
					ERROR_INCORRECT_TYPE_TITLE, 
					JOptionPane.OK_CANCEL_OPTION, 
					JOptionPane.QUESTION_MESSAGE);
			
			// If OK, the process is launched. Otherwise, it is not.
			if (result == JOptionPane.OK_OPTION) {
				orientations = new double[collection.size()];
				orientationSegments(layer, collection);
			} else {
				return false;
			}
		}
		
		// Histogram
		if (dialog.getCheckBox(HISTOGRAM).isSelected()) {
			Histogram h = new Histogram(
					TITLE_HISTOGRAM,
					orientations,
					dialog.getInteger(HISTOGRAM_STEP),
					"orientations");
			h.pack();
			RefineryUtilities.centerFrameOnScreen(h);
			h.setVisible(true);
		}
		
		return true;
	}

	
	/** 
	 * Compute and add several fields in the {@link Layer} of segments (do not compute
	 * for {@link Geometry} other than a segment): length of the segment, orientation,
	 * and a classification of the orientations. NaN is used for non segments.
	 * 
	 * @param layer
	 * @param collection
	 * @return true
	 */
	@SuppressWarnings("unchecked")
	private boolean orientationSegments(Layer layer, FeatureCollection collection) {

		int unit = jComboUnits.getSelectedIndex();
		int interval = jComboInterval.getSelectedIndex();
		boolean accuracy = dialog.getCheckBox(ACCURACY).isSelected();
		boolean histogram = dialog.getCheckBox(HISTOGRAM).isSelected();
		boolean classification = dialog.getCheckBox(CLASSIFICATION).isSelected();
		
		// Step of the classification
		double classifStep = dialog.getDouble(CLASSIFICATION_STEP);
		if (classifStep <= 0) {
			classifStep = 1.0;
		}
		
		// Add attributes "length", "orientation" (reduced to "orientatio" to fit with
		// the maximum attribute length in ESRI Shapefile)
		// and "classe" (classification) in the schema
		FeatureSchema fsNew = (FeatureSchema) collection.getFeatureSchema().clone();
		fsNew.addAttribute("length", AttributeType.DOUBLE);
		fsNew.addAttribute("orientatio", AttributeType.DOUBLE);
		if (classification) {
			fsNew.addAttribute("classe", AttributeType.INTEGER);
		}			
		
		FeatureDataset newCollection = new FeatureDataset(fsNew);		
		
		// All features are copied to avoid the use of the setSchema()
		// (too low level)
		int k = 0;
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature feature = it.next();
			
			Feature newFeature = new BasicFeature(fsNew);
			for (int i = 0 ; i < feature.getAttributes().length ; i++) {
				newFeature.setAttribute(i, feature.getAttribute(i));
			}
			
			Geometry geom = feature.getGeometry();
			Coordinate[] coord = geom.getCoordinates();
			
			if ((geom instanceof LineString || (geom instanceof MultiLineString && geom.getNumGeometries() ==1))
					&& coord.length == 2
					&& !GeometryUtil.compareCoordinate(coord[0], coord[1])) {
				
				LineString line = (LineString) feature.getGeometry();

				// Add the value of the attribute "length"
				newFeature.setAttribute("length", line.getLength());			

				Double orientation = new Double(
						AngleUtil.angle(line.getStartPoint(),line.getEndPoint(), 
										unit, interval, accuracy));
			
				// Add the value of the attribute "orientatio" 
				newFeature.setAttribute("orientatio", orientation);

				// Add the value of the attribute classification if required
				if (classification) {
					int classe = new Double(orientation / classifStep).intValue();
					newFeature.setAttribute("classe", classe);
				}
				
				newCollection.add(newFeature);

				// Add the value of the orientations to build an histogram if required
				if (histogram) {
					orientations[k] = orientation.doubleValue();
					k++;				
				}				
			} else { // It is not a segment: not computed				
				LOGGER.warn("Feature ID = " + feature.getID() + " is not process: not a segment.");
			}
		}
		
        layer.setFeatureCollection(newCollection);
        layer.fireLayerChanged(LayerEventType.METADATA_CHANGED);

		return true;
	}

}