/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.plugin.polygon;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JOptionPane;
import javax.swing.JTextField;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.index.SpatialIndex;
import com.vividsolutions.jts.index.strtree.STRtree;
import com.vividsolutions.jump.feature.AttributeType;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.feature.IndexedFeatureCollection;
import com.vividsolutions.jump.task.TaskMonitor;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.LayerEventType;
import com.vividsolutions.jump.workbench.model.StandardCategoryNames;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.plugin.ThreadedPlugIn;
import com.vividsolutions.jump.workbench.ui.ErrorDialog;
import com.vividsolutions.jump.workbench.ui.GUIUtil;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;
import com.vividsolutions.jump.workbench.ui.plugin.FeatureInstaller;

import fr.ign.cogit.alpage.morphal.algorithm.maths.Combinatorics;
import fr.ign.cogit.alpage.morphal.geometry.PolygonUtil;
import fr.ign.cogit.alpage.morphal.i18n.I18NPlug;

/** 
 * Plugin to characterise a {@link Layer} of {@link Polygon}s (rectangles - 3 levels -, circles)
 *
 * @author Eric Grosso - IGN / Laboratoire COGIT
 * 
 * @version 0.1 (2011-01-09)
 */
public class PolygonCharacterisationPlugin extends AbstractPlugIn implements ThreadedPlugIn {
	
	////////////////////////////////////////////////////////////////////////////////////
	// Internationalisation
	
	private final static String MENU = I18NPlug.getI18N("project");
	private final static String MENU_SUBTITLE = I18NPlug.getI18N("polygon");
	private final static String TITLE = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.title");
	private final static String DESCRIPTION = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.description");
	
	private final static String LAYER = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.layer");
	private final static String LAYER_URBAN_BLOCK = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.layerUrbanBlock");

	// Errors
	private final static String ERROR = I18NPlug.getI18N("error");
	private final static String EMPTY = I18NPlug.getI18N("error.emptyCollection");
	private final static String IDENTICAL_LAYERS = I18NPlug.getI18N("error.identicalLayers");
	private final static String ERROR_INCORRECT_TYPE_TITLE = I18NPlug.getI18N("error.polygon.incorrectGeometricTypeTitle");
	private final static String ERROR_INCORRECT_TYPE = I18NPlug.getI18N("error.polygon.incorrectGeometricType");
	private final static String ERROR_MULTIPOLYGON_TYPE = I18NPlug.getI18N("error.polygon.multiPolygonType");
	private final static String ERROR_INCORRECT_TYPE_ALL = I18NPlug.getI18N("error.polygon.incorrectGeometricTypeAll");

	// Add fields or/and layers
	private final static String CHANGES = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.changes");
//	private final static String ADD_FIELDS = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.fields");
	private final static String ADD_LAYERS = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.layers");

	// Rectangle
	private final static String RECTANGLE = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.rectangle");
	private final static String RECTANGLE_L1 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.detectionRectangleL1");
	private final static String RECTANGLE_L2 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.detectionRectangleL2");
	private final static String RECTANGLE_L3 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.detectionRectangleL3");
	private final static String RECTANGLE_GROUP = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.detectionRectangularGroup");

	private final static String SD_CONVEX_1 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceConvexL1");
	private final static String SD_CONVEX_2 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceConvexL2");
	private final static String SD_CONVEX_3 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceConvexL3");
	private final static String SD_CONVEX_GROUP = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceConvexGroup");
	private final static String SD_MBR_1 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceMBRL1");
	private final static String SD_MBR_2 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceMBRL2");
	private final static String SD_MBR_3 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceMBRL3");
	private final static String SD_MBR_GROUP = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.surfaceDistanceMBRGroup");

	private final static String ORIENTATION = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.orientation");
	private final static String ORIENTATION_NOT_COMPUTED = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.orientationNotComputed");
	private final static String ORIENTATION_METHOD_MBR = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.orientationMethodMBR");
	//private final static String ORIENTATION_METHOD_2 = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.orientationMethod2");

	// Circle
	private final static String CIRCLE = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.circle");
	private final static String CIRCLE_DETECTION = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.detectionCircle");
	private final static String MILLER = I18NPlug.getI18N("polygon.PolygonCharacterisationPlugin.miller");

	////////////////////////////////////////////////////////////////////////////////////

	private static MultiInputDialog dialog;

	// Rectangle
	private JCheckBox jcbRectangleL1, jcbRectangleL2, jcbRectangleLevel2, jcbRectangleGoup;
	private JComboBox jComboRectangle, jComboLayer;
	private JTextField sdConvexL1, sdConvexL2, sdConvexL3, sdConvexGroup; // surface distance -- convex
	private JTextField sdMBRL1, sdMBRL2, sdMBRL3, sdMBRGroup; // surface distance -- MBR

	// Circle
	private JCheckBox jcbCircle;
	private JTextField miller;

	// Collection of rectangular, non rectangular or circular polygons
	private FeatureDataset collectionNonRectangle = null, collectionRectangleL1 = null, collectionRectangleL2 = null, 
		collectionRectangleL3 = null, collectionRectangleGroup = null, collectionNonRectangleGroup = null,
		collectionCircle = null;
	
	private List<Integer> idOfNonRectangle = new ArrayList<Integer>();


	/** 
	 * Initialisation.
	 * 
	 * @param context
	 */
	@Override
	public void initialize(PlugInContext context) {
		FeatureInstaller featureInstaller = new FeatureInstaller(context.getWorkbenchContext());
		featureInstaller.addMainMenuItem(
				this,
				new String[] { MENU, MENU_SUBTITLE },
				TITLE,
				false,
				null,
				new MultiEnableCheck()
				.add(context.getCheckFactory().createTaskWindowMustBeActiveCheck())
				.add(context.getCheckFactory().createAtLeastNLayersMustExistCheck(1)));
	}

	
	/** 
	 * Global execution.
	 * 
	 * @param context
	 * @return true if process is executed, otherwise false
	 */
	@Override
	public boolean execute(PlugInContext context) {
		this.reportNothingToUndoYet(context);
		dialog = new MultiInputDialog(context.getWorkbenchFrame(), getName(), true);
		setDialogValues(dialog, context);
		GUIUtil.centreOnWindow(dialog);
		dialog.setVisible(true);
		if (! dialog.wasOKPressed()) { return false; }
		return true;
	}

	
	/** 
	 * Initialisation of the dialog interface.
	 * 
	 * @param dialog
	 * @param context
	 */
	private void setDialogValues(MultiInputDialog dialog, PlugInContext context) {

		dialog.setSideBarDescription(DESCRIPTION);

		dialog.addLayerComboBox(LAYER,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());

		// Rectangular polygons: characterisation level 1
		dialog.addSeparator();
		dialog.addLabel(RECTANGLE);
		dialog.addSeparator();
		this.jcbRectangleL1 = dialog.addCheckBox(RECTANGLE_L1, true);
		this.jcbRectangleL1.addChangeListener(new RectangleChange());
		this.sdConvexL1 = dialog.addDoubleField(SD_CONVEX_1, 0.05, 4, SD_CONVEX_1);
		this.sdMBRL1 = dialog.addDoubleField(SD_MBR_1, 0.05, 4, SD_MBR_1);

		// Rectangular polygons: characterisation level 2
		dialog.addSeparator();
		this.jcbRectangleL2 = dialog.addCheckBox(RECTANGLE_L2, false);
		this.sdConvexL2 = dialog.addDoubleField(SD_CONVEX_2, 0.1, 4, SD_CONVEX_2);
		this.sdConvexL2.setEnabled(false);
		this.sdMBRL2 = dialog.addDoubleField(SD_MBR_2, 0.1, 4, SD_MBR_2);
		this.sdMBRL2.setEnabled(false);
		this.jcbRectangleL2.addChangeListener(new RectangleLevelTwoChange());

		// Rectangular polygons: characterisation level 3
		dialog.addSeparator();
		this.jcbRectangleLevel2 = dialog.addCheckBox(RECTANGLE_L3, false);
		this.sdConvexL3 = dialog.addDoubleField(SD_CONVEX_3, 0.15, 4, SD_CONVEX_3);
		this.sdConvexL3.setEnabled(false);
		this.sdMBRL3 = dialog.addDoubleField(SD_MBR_3, 0.15, 4, SD_MBR_3);
		this.sdMBRL3.setEnabled(false);
		this.jcbRectangleLevel2.setEnabled(false);
		this.jcbRectangleLevel2.addChangeListener(new RectangleLevelThreeChange());

		// Rectangular polygons: characterisation group level
		dialog.addSeparator();
		this.jcbRectangleGoup = dialog.addCheckBox(RECTANGLE_GROUP, false);
		this.jcbRectangleGoup.setEnabled(true);
		this.jcbRectangleGoup.addChangeListener(new RectangleGroupRectChange());
		this.jComboLayer = dialog.addLayerComboBox(LAYER_URBAN_BLOCK,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());
		this.jComboLayer.setEnabled(false);
		this.sdConvexGroup = dialog.addDoubleField(SD_CONVEX_GROUP, 0.1, 4, SD_CONVEX_GROUP);
		this.sdConvexGroup.setEnabled(false);
		this.sdMBRGroup = dialog.addDoubleField(SD_MBR_GROUP, 0.1, 4, SD_MBR_GROUP);
		this.sdMBRGroup.setEnabled(false);

		// Orientation
		dialog.addSeparator();
		Collection<String> items = new ArrayList<String>();
		items.add(ORIENTATION_NOT_COMPUTED);
		items.add(ORIENTATION_METHOD_MBR);
		//items.add(ORIENTATION_METHOD_2);
		this.jComboRectangle = dialog.addComboBox(ORIENTATION, ORIENTATION_NOT_COMPUTED, items, "");


		// Circular polygons: characterisation
		dialog.addSeparator();
		dialog.addLabel(CIRCLE);
		dialog.addSeparator();

		this.jcbCircle = dialog.addCheckBox(CIRCLE_DETECTION, true);
		this.miller = dialog.addDoubleField(MILLER, 0.95, 4, MILLER);
		this.jcbCircle.addChangeListener(new CercleChange());

		// Management of layers and fields in outputs
		dialog.addSeparator();
		dialog.addLabel(CHANGES);
		dialog.addSeparator();
		//dialog.addCheckBox(ADD_FIELDS, true);
		dialog.addCheckBox(ADD_LAYERS, false);
	}


	/** Internal class: management of the "rectangle" dialog: level 1. */
	class RectangleChange implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			boolean flag = jcbRectangleL1.isSelected();
			boolean flagLTwo = jcbRectangleL2.isSelected();
			boolean flagLThree = jcbRectangleLevel2.isSelected();
			boolean flagGroup = jcbRectangleGoup.isSelected();

			if (!flag) {
				sdConvexL1.setEnabled(flag);
				sdMBRL1.setEnabled(flag);

				jcbRectangleL2.setSelected(flag);
				jcbRectangleL2.setEnabled(flag);
				sdConvexL2.setEnabled(flag);
				sdMBRL2.setEnabled(flag);

				jcbRectangleLevel2.setSelected(flag);
				jcbRectangleLevel2.setEnabled(flag);
				sdConvexL3.setEnabled(flag);
				sdMBRL3.setEnabled(flag);

				jcbRectangleGoup.setSelected(flag);
				jcbRectangleGoup.setEnabled(flag);
				sdConvexGroup.setEnabled(flag);
				sdMBRGroup.setEnabled(flag);

				jComboRectangle.setEnabled(flag);				
			} else {
				sdConvexL1.setEnabled(flag);
				sdMBRL1.setEnabled(flag);

				jcbRectangleL2.setEnabled(flag);
				sdConvexL2.setEnabled(flagLTwo);
				sdMBRL2.setEnabled(flagLTwo);

				jcbRectangleLevel2.setEnabled(flag && flagLTwo);
				sdConvexL3.setEnabled(flagLThree && flagLTwo);
				sdMBRL3.setEnabled(flagLThree && flagLTwo);

				jcbRectangleGoup.setEnabled(flag);
				sdConvexGroup.setEnabled(flagGroup);
				sdMBRGroup.setEnabled(flagGroup);

				jComboRectangle.setEnabled(flag);
			}
		}
	}


	/** Internal class: management of the "rectangle" dialog: level 2. */
	class RectangleLevelTwoChange implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			boolean flag = jcbRectangleL2.isSelected();
			sdConvexL2.setEnabled(flag);
			sdMBRL2.setEnabled(flag);

			// Level 3 management
			jcbRectangleLevel2.setEnabled(flag);
			if (!flag && jcbRectangleLevel2.isSelected()) {				
				jcbRectangleLevel2.setSelected(flag);
			}
		}
	}


	/** Internal class: management of the "rectangle" dialog: level 3. */
	class RectangleLevelThreeChange implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			boolean flag = jcbRectangleLevel2.isSelected();
			sdConvexL3.setEnabled(flag);
			sdMBRL3.setEnabled(flag);
		}
	}


	/** Internal class: management of the "rectangle" dialog: group level. */
	class RectangleGroupRectChange implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			boolean flag = jcbRectangleGoup.isSelected();
			jComboLayer.setEnabled(flag);
			sdConvexGroup.setEnabled(flag);
			sdMBRGroup.setEnabled(flag);
		}
	}


	/** Internal classe: management of the "circle" dialog. */
	class CercleChange implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			boolean flag = jcbCircle.isSelected();
			miller.setEnabled(flag);
		}
	}

	
	/** 
	 * Execution.
	 * 
	 * @param monitor
	 * @param context
	 */
	public void run(TaskMonitor monitor, PlugInContext context) {
		characterisationPolygon(context);
		System.gc();
	}


	/** 
	 * Characterisation of {@link Polygon}s.
	 * 
	 * @param context
	 * @return true if the process can be launched, otherwise false
	 */
	@SuppressWarnings("unchecked")
	private boolean characterisationPolygon(PlugInContext context) {

		// Layer to process
		Layer layer = dialog.getLayer(LAYER);

		// Features of the layer to process
		final FeatureCollection collection = layer.getFeatureCollectionWrapper();

		if (collection.isEmpty()) {
			ErrorDialog.show(context.getWorkbenchFrame(),
					ERROR,
					EMPTY,
					"");
			return false; 
		}

		// Management of incorrect geometric input (other than Polygon or MultiPolygon)
		// This preprocess is not long and allows to users to make a choice about what they
		// want to do.
		int nP = 0; // Number of features with a Polygon geometric type
		int nMP = 0; // Number of features with a MultiPolygon geometric type
		
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature f = it.next();
			if (f.getGeometry() instanceof Polygon) nP++;
			else if (f.getGeometry() instanceof MultiPolygon) nMP++;
		}
		
		int size = collection.size();
		int incorrectType = size - (nP + nMP);
		
		if (nP == size) { // All features have a correct type
			characterisationPolygon(context, collection);
		} else if (incorrectType == size) { // All features have an incorrect type
			ErrorDialog.show(
					context.getWorkbenchFrame(),
					ERROR_INCORRECT_TYPE_TITLE,
					ERROR_INCORRECT_TYPE_ALL,
					"");
			return false;
		} else if (nMP != 0) {
			if (incorrectType == 0) { // Cases of multipolygons only
				JOptionPane.showMessageDialog(
						context.getWorkbenchFrame(),
						nMP + " " + ERROR_MULTIPOLYGON_TYPE,
						ERROR_INCORRECT_TYPE_TITLE,
						JOptionPane.OK_OPTION);

				characterisationPolygon(context, collection);				
			} else { // Cases of multipolygons and incorrect geometric type
				int result = JOptionPane.showConfirmDialog(
						context.getWorkbenchFrame(),
						nMP + " " + ERROR_MULTIPOLYGON_TYPE +
						"\n" + incorrectType + " " +  ERROR_INCORRECT_TYPE,
						ERROR_INCORRECT_TYPE_TITLE,
						JOptionPane.OK_CANCEL_OPTION,
						JOptionPane.QUESTION_MESSAGE);

				// If OK, the characterisation process is launched. Otherwise, it is not.
				if (result == JOptionPane.OK_OPTION) {
					characterisationPolygon(context, collection);
				} else {
					return false;
				}
			}
		}
		return true;
	}


	/** 
	 * Characterisation of polygons
	 * 
	 * @param context
	 * @param collection
	 * @return boolean
	 */
	@SuppressWarnings("unchecked")
	private boolean characterisationPolygon(PlugInContext context, FeatureCollection collection) {

		// Level 1
		boolean rectangle = jcbRectangleL1.isSelected();
		double sdConvL1 = dialog.getDouble(SD_CONVEX_1);
		double sdMBRectL1 = dialog.getDouble(SD_MBR_1);

		// Level 2
		boolean rectangleTwo = jcbRectangleL2.isSelected();
		double sdConvL2 = dialog.getDouble(SD_CONVEX_2);
		double sdMBRectL2 = dialog.getDouble(SD_MBR_2);

		// Level 3
		boolean rectangleThree = jcbRectangleLevel2.isSelected();
		double sdConvL3 = dialog.getDouble(SD_CONVEX_3);
		double sdMBRectL3 = dialog.getDouble(SD_MBR_3);

		// Level Group
		boolean rectangleGroup = jcbRectangleGoup.isSelected();
		double sdConvGroup = dialog.getDouble(SD_CONVEX_GROUP);
		double sdMBRectGroup = dialog.getDouble(SD_MBR_GROUP);

		// Orientation
//		int indexCombo = dialog.getComboBox(ORIENTATION).getSelectedIndex();		

		// Circle
		boolean circle = jcbCircle.isSelected();
		double iMiller = dialog.getDouble(MILLER);

		// Adding: fields / layers
//		boolean fields = dialog.getCheckBox(ADD_FIELDS).isSelected();
		boolean layers = dialog.getCheckBox(ADD_LAYERS).isSelected();

		if (rectangle == false && circle == false) return false;

		Layer layer = dialog.getLayer(LAYER);

		FeatureDataset newCollection = characterisationPolygon(collection,
				rectangle, sdConvL1, sdMBRectL1,
				rectangleTwo, sdConvL2, sdMBRectL2,
				rectangleThree, sdConvL3, sdMBRectL3,
				circle, iMiller);

		// Add fields
//		if(fields) {
		layer.setFeatureCollection(newCollection);
		layer.fireLayerChanged(LayerEventType.METADATA_CHANGED);
//		}

		// Add layers
		if (layers) {

			// Rectangular layers
			if (rectangle) {
				// Modified schema
				FeatureSchema fsNew = (FeatureSchema) newCollection.getFeatureSchema().clone();

				collectionNonRectangle = new FeatureDataset(fsNew);
				collectionRectangleL1 = new FeatureDataset(fsNew);
				collectionRectangleL2 = new FeatureDataset(fsNew);
				collectionRectangleL3 = new FeatureDataset(fsNew);

				int index = 0;
				for (Iterator<Feature> it = newCollection.iterator() ; it.hasNext();) {
					Feature feature = it.next();
					Feature newFeature = new BasicFeature(fsNew);
					for (int i = 0 ; i < feature.getAttributes().length ; i++) {
						newFeature.setAttribute(i, feature.getAttribute(i));
					}

					int rect = ((Integer)newFeature.getAttribute("rectangle")).intValue();

					if (rect == 0) {
						collectionNonRectangle.add(newFeature);
						// This add allows to categorise rectangular groups with a level 4
						idOfNonRectangle.add(index);
					} else if (rect == 1) {
						collectionRectangleL1.add(newFeature);
					} else if (rect == 2) {
						collectionRectangleL2.add(newFeature);
					} else { // Case where rect = 3
						collectionRectangleL3.add(newFeature);
					}
					index++;
				}
			}

			// Circular layer
			if (circle) {
				// Modified schema
				FeatureSchema fsNew = (FeatureSchema) newCollection.getFeatureSchema().clone();

				collectionCircle = new FeatureDataset(fsNew);
			}

		}

		///////////////////////////////////////////////////////////
		////////// CHARACTERISATION OF RECTANGULAR GROUPS /////////
		///////////////////////////////////////////////////////////

		// Detection of rectangular groups
		if (rectangleGroup) {
			Layer layerUrbanBlock = dialog.getLayer(LAYER_URBAN_BLOCK);

			// Stop the process if the selected layers are identical
			if (layer.getName().equals(layerUrbanBlock.getName())) {
				ErrorDialog.show(context.getWorkbenchFrame(),
						ERROR,
						IDENTICAL_LAYERS,
						"");
				return false;
			}

			// Modified schema
			FeatureSchema fsNew = (FeatureSchema) newCollection.getFeatureSchema().clone();

			// Creation of the dataset collectionNonRectangle if it has not been created before
			// (in the case where layers have not to be displayed)
			if (!layers) {

				collectionNonRectangle = new FeatureDataset(fsNew);

				for (Iterator<Feature> it = newCollection.iterator() ; it.hasNext();) {
					Feature feature = it.next();
					Feature newFeature = new BasicFeature(fsNew);
					for (int i = 0 ; i < feature.getAttributes().length ; i++) {
						newFeature.setAttribute(i, feature.getAttribute(i));
					}

					int rect = ((Integer)newFeature.getAttribute("rectangle")).intValue();

					if (rect == 0) {
						collectionNonRectangle.add(newFeature);
					}
				}
			}

			// Creation of the layer in link with the rectangular groups
			collectionRectangleGroup = new FeatureDataset(fsNew);

			// Spatial indexes
			SpatialIndex s = new STRtree(10);
			SpatialIndex s2 = new STRtree(10);

			// Indexation of urban blocks
			final FeatureCollection collectionIlots = layerUrbanBlock.getFeatureCollectionWrapper();
			IndexedFeatureCollection indexedCollectionUrbanBlocks = new IndexedFeatureCollection(collectionIlots,s);

			// Indexation of parcels
			IndexedFeatureCollection indexedCollectionParcels = new IndexedFeatureCollection(newCollection,s2);

			// Parcels which will be considered as rectangles in association with other parcels in their neighbourhood 
			List<Integer> parcelsRectangleInGroup = new ArrayList<Integer>();
			
			int idObjet = 0;
			// For each non rectangular parcel, find its associated urban block
			for (Feature fNonRect : (List<Feature>)collectionNonRectangle.getFeatures()) {
				List<Feature> candidatesUrbanBlock = indexedCollectionUrbanBlocks.query(fNonRect.getGeometry().getEnvelopeInternal());

				Feature urbanBlock = null;

				for (Feature c : candidatesUrbanBlock) {
					// If the tested urban block matches
					if (c.getGeometry().contains(fNonRect.getGeometry())) {
						urbanBlock = c;
					}
				}

				// If at least one urban block matches, creation of the list of all features around the considered non rectangular parcel
				if (urbanBlock != null) {
					List<Feature> candidatesParcels = indexedCollectionParcels.query(urbanBlock.getGeometry().getEnvelopeInternal());
					List<Feature> parcelsAdjacent = new ArrayList<Feature>();

					for (Feature parcel : candidatesParcels) {
						// Using the "touches" method allows to not proces the considered parcel as an adjacent parcel
						// (fNonRect), due to the fact that it doesn't "touch itself
						if (urbanBlock.getGeometry().contains(parcel.getGeometry())
								&& parcel.getGeometry().touches(fNonRect.getGeometry())) {
							parcelsAdjacent.add(parcel);
						}
					}

					// If the considered and non rectangular parcel has adjacent parcels,
					// the considerd parcel is grouped with adajacent parcels to check
					// if the group is rectangle or not.
					// TODO: The second condition of the following test has to be changed.
					// It is not currently possible due to a huge problem to allocate more memory.
					// Nevertheless, this condition is efficient to process the majority of cases.
					if (!parcelsAdjacent.isEmpty() && parcelsAdjacent.size() <= 20) {
						
						// Generation of all possible combinations without repetitions for adjacent parcels
						List<List<Integer>> combinations = Combinatorics.combination(parcelsAdjacent.size());

						// Max of index: square_root(surfaceDistanceConvex^2 + surfaceDistanceMBR^2) = sqrt(2) 
						double indexMin = 2;
						double sdConvexMinGroup = 1;
						double sdMBRMinGroup = 1;

						Polygon mbrMinG = null;

						for (int k = 0 ; k < combinations.size() ; k++) {
							List<Integer> combination = combinations.get(k);

							// Geometry initialisation
							Geometry geometryTest = fNonRect.getGeometry();						

							for (int i = 0 ; i < combination.size() ; i++) {
								// The resulting geometry is not necessarily a polygon
								// (if the intersection is only a single point, it isn't a polygon but a multipolygon)
								geometryTest = geometryTest.union(parcelsAdjacent.get(combination.get(i)).getGeometry());
							}
							
							// Test to remove cases where the intersection is a single point:
							// This particular case can not be a rectangle
							if (geometryTest.getNumGeometries() == 1) {
								// Test if the group of geometries is rectangular or not.
								// Computation of indices to detect rectangle shapes
								Polygon convexHull = (Polygon) geometryTest.convexHull();
								Polygon mbrMin = PolygonUtil.MBR((Polygon)geometryTest);

								double distanceSurfaceConvex = PolygonUtil.surfaceDistance(geometryTest, convexHull);
								double distanceSurfaceMBR = PolygonUtil.surfaceDistance(geometryTest, mbrMin);

								// Test to find the optimal combination
								double indiceTemp = Math.pow(distanceSurfaceConvex,2) + Math.pow(distanceSurfaceMBR,2);
								if (indiceTemp < indexMin) {
									indexMin = indiceTemp;
									sdConvexMinGroup = distanceSurfaceConvex;
									sdMBRMinGroup = distanceSurfaceMBR;
									mbrMinG = mbrMin;
								}								
							}
						}

						// If a solution is found (it is a priori the case, because the surface distance
						// based on the convex hull has always a value stricly less than 1)
						if (indexMin != -1) {
							double orientation = PolygonUtil.isRectangle(sdConvexMinGroup, sdConvGroup,
									sdMBRMinGroup, sdMBRectGroup, mbrMinG);

							// If the rectangle satisfies the chosen indices in relationship
							// with the group level characterisation
							if(orientation != -1) {								
								Feature newFeature = new BasicFeature(fsNew);
								for (int i = 0 ; i < fNonRect.getAttributes().length ; i++) {
									newFeature.setAttribute(i, fNonRect.getAttribute(i));	
								}

								newFeature.setAttribute("sd_convex", sdConvexMinGroup);
								newFeature.setAttribute("sd_mbr", sdMBRMinGroup);
								newFeature.setAttribute("rectangle", new Integer(4));
								newFeature.setAttribute("orient_rec", new Double(orientation));
								// TODO: look if it is not better to provide the elongation of the polygon itself
								// rather than the one of the group
								newFeature.setAttribute("elongation",
										new Double(PolygonUtil.elongation(mbrMinG)));

								collectionRectangleGroup.add(newFeature);
								parcelsRectangleInGroup.add(0,idObjet);
								newCollection.getFeature(idOfNonRectangle.get(idObjet)).setAttribute("rectangle",4);
							}
						}
					}
				}
				idObjet++;
			}
			
			// Creation of a new layer which contains the final non rectangular polygons
			// after the rectangular group characterisation
			// Copy first of the layer which contains the non rectangular polygons
			// before the rectangular group characterisation
			if (!parcelsRectangleInGroup.isEmpty()) {
				collectionNonRectangleGroup = new FeatureDataset(fsNew);

				for (Feature fNonRect : (List<Feature>)collectionNonRectangle.getFeatures()) {
					Feature newFeature = new BasicFeature(fsNew);
					for (int i = 0 ; i < fNonRect.getAttributes().length ; i++) {
						newFeature.setAttribute(i, fNonRect.getAttribute(i));	
					}
					collectionNonRectangleGroup.add(newFeature);
				}

				// Remove the rectangular group polygons
				for (int i = 0 ; i < parcelsRectangleInGroup.size() ; i++) {
					collectionNonRectangleGroup.remove(collectionNonRectangleGroup.getFeature(parcelsRectangleInGroup.get(i).intValue()));
				}
			}
		}

		///////////////////////////////////////////////////////////
		//////// CHARACTERISATION OF RECTANGULAR GROUPS: END //////
		///////////////////////////////////////////////////////////

		
		// Display the different computed layers, category by category
		if (layers) {

			// Display the layers of rectangular and non rectangular features
			if (rectangle) {
				if (rectangleGroup) {
					if (!collectionNonRectangleGroup.isEmpty()) {
						context.addLayer(StandardCategoryNames.RESULT, "non_rectangle_after_group", collectionNonRectangleGroup);
					}
					if (!collectionRectangleGroup.isEmpty()) {
						context.addLayer(StandardCategoryNames.RESULT, "rectangle_group", collectionRectangleGroup);
					}
				}
				if (!collectionNonRectangle.isEmpty()) {
					context.addLayer(StandardCategoryNames.RESULT, "non_rectangle", collectionNonRectangle);
				}
				if (!collectionRectangleL3.isEmpty()) {
					context.addLayer(StandardCategoryNames.RESULT, "rectangle_level_3", collectionRectangleL3);
				}
				if (!collectionRectangleL2.isEmpty()) {
					context.addLayer(StandardCategoryNames.RESULT, "rectangle_level_2", collectionRectangleL2);
				}
				if (!collectionRectangleL1.isEmpty()) {
					context.addLayer(StandardCategoryNames.RESULT, "rectangle_level_1", collectionRectangleL1);
				}
			}

			// Display the layer of circular features
			if (!collectionCircle.isEmpty()) {
				context.addLayer(StandardCategoryNames.RESULT, "circle", collectionCircle);
			}	
		}

		return true;
	}


	/** 
	 * Characterisation of {@link Polygon}s: rectangles (3 different possible levels),
	 * circle, and management of the data schema.
	 * 
	 * @param collection collection of polyongs to process
	 * @param rectangleDetectionL1 true if the detection of rectangular shapes (level 1) is activated, otherwise false
	 * @param sdConvL1 threshold of surface distance between the polygone and its convex hull (level 1)
	 * @param sdMBRLevel1 threshold of surface distance between the polygone and its MBR (level 1)
	 * @param rectangleDetectionL2 true if the detection of rectangular shapes (level 2) is activated, otherwise false
	 * @param sdConvL2 threshold of surface distance between the polygone and its convex hull (level 2)
	 * @param sdMBRLevel2 sdMBRLevel1 threshold of surface distance between the polygone and its MBR (level 2)
	 * @param rectangleDetectionL3 true if the detection of rectangular shapes (level 3) is activated, otherwise false
	 * @param sdConvL3 threshold of surface distance between the polygone and its convex hull (level 3)
	 * @param sdMBRLevel1 threshold of surface distance between the polygone and its MBR (level 3)
	 * @param circleDetection true if the detection of circular shapes, otherwise false
	 * @param iMiller minimum threshold Miller index to characterise a polygon as a circle
	 * @return a dataset of characterised polygons
	 */
	@SuppressWarnings("unchecked")
	private FeatureDataset characterisationPolygon(FeatureCollection collection,
			boolean rectangleDetectionL1, double sdConvL1, double sdMBRLevel1,
			boolean rectangleDetectionL2, double sdConvL2, double sdMBRLevel2,
			boolean rectangleDetectionL3, double sdConvL3, double sdMBRLevel3,			
			boolean circleDetection, double iMiller) {

		FeatureSchema fsNew = (FeatureSchema) collection.getFeatureSchema().clone();

		// Modification of the original schema if the "rectangle" detection is activated
		if (rectangleDetectionL1) {
			fsNew.addAttribute("rectangle", AttributeType.INTEGER);
			fsNew.addAttribute("orient_rec", AttributeType.DOUBLE);
			fsNew.addAttribute("elongation", AttributeType.DOUBLE);
			fsNew.addAttribute("sd_convex", AttributeType.DOUBLE);
			fsNew.addAttribute("sd_mbr", AttributeType.DOUBLE);
		}

		// Modification of the original schema if the "circle" detection is activated
		if (circleDetection) {
			fsNew.addAttribute("circle", AttributeType.INTEGER);
		}

		// Modification of the original schema if the "multipolygon" detection is activated
		fsNew.addAttribute("multipolygon", AttributeType.STRING);

		FeatureDataset newCollection = new FeatureDataset(fsNew);

		// If the "rectangle" detection level 1 is activated
		// Remark: Separation with the "circle" detection for optimisation only
		if (rectangleDetectionL1) {
			for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
				Feature feature = it.next();
				Feature newFeature = new BasicFeature(fsNew);
				for (int i = 0 ; i < feature.getAttributes().length ; i++) {
					newFeature.setAttribute(i, feature.getAttribute(i));
				}

				boolean flagMulti = false;

				Polygon polygon = null;
				if (feature.getGeometry() instanceof Polygon) {
					polygon = (Polygon)feature.getGeometry();
				}
				else if (feature.getGeometry() instanceof MultiPolygon) {
					MultiPolygon multipolygon = (MultiPolygon)feature.getGeometry();

					// Choice of the polygon which has the biggest area in the multipolygon
					double areaMax = Double.MIN_VALUE;
					for (int i = 0 ; i < multipolygon.getNumGeometries() ; i++) {
						Polygon p = (Polygon)multipolygon.getGeometryN(i);
						if (p.getArea() > areaMax) {
							polygon = p;
						}
					}
					flagMulti = true;
				}

				// Computation of indices to detect rectangle shapes
				Polygon convexHull = (Polygon) polygon.convexHull();
				Polygon mbrMin = PolygonUtil.MBR(polygon);

				double surfaceDistConvex = PolygonUtil.surfaceDistance(polygon, convexHull);
				double surfaciqueDistMBR = PolygonUtil.surfaceDistance(polygon, mbrMin);

				double orientation = PolygonUtil.isRectangle(surfaceDistConvex, sdConvL1,
						surfaciqueDistMBR, sdMBRLevel1, mbrMin);
				// If rectangle level 1
				if(orientation != -1.0) {
					newFeature.setAttribute("rectangle", new Integer(1));
					newFeature.setAttribute("orient_rec", new Double(orientation));
					newFeature.setAttribute("elongation",
							new Double(PolygonUtil.elongation(mbrMin)));
					newFeature.setAttribute("sd_convex", surfaceDistConvex);
					newFeature.setAttribute("sd_mbr", surfaciqueDistMBR);
				} else {
					newFeature.setAttribute("rectangle", new Integer(0));
					newFeature.setAttribute("orient_rec", new Double(0));
					newFeature.setAttribute("elongation",
							new Double(PolygonUtil.elongation(mbrMin)));
					newFeature.setAttribute("sd_convex", surfaceDistConvex);
					newFeature.setAttribute("sd_mbr", surfaciqueDistMBR);

					// If the "rectangle" detection level 2 is activated
					if (rectangleDetectionL2) {
						double orientationTwo = PolygonUtil.isRectangle(surfaceDistConvex, sdConvL2,
								surfaciqueDistMBR, sdMBRLevel2, mbrMin);
						// If rectangle level 2
						if(orientationTwo != -1.0) {
							newFeature.setAttribute("rectangle", new Integer(2));
							newFeature.setAttribute("orient_rec", new Double(orientationTwo));
							newFeature.setAttribute("elongation",
									new Double(PolygonUtil.elongation(mbrMin)));
							newFeature.setAttribute("sd_convex", surfaceDistConvex);
							newFeature.setAttribute("sd_mbr", surfaciqueDistMBR);
						} else {
							// If the "rectangle" detection of level 3 is activated
							if (rectangleDetectionL3) {
								double orientationThree = PolygonUtil.isRectangle(surfaceDistConvex, sdConvL3,
										surfaciqueDistMBR, sdMBRLevel3, mbrMin);
								// If rectangle level 3
								if(orientationThree != -1.0) {
									newFeature.setAttribute("rectangle", new Integer(3));
									newFeature.setAttribute("orient_rec", new Double(orientationThree));
									newFeature.setAttribute("elongation",
											new Double(PolygonUtil.elongation(mbrMin)));
								}
								newFeature.setAttribute("sd_convex", surfaceDistConvex);
								newFeature.setAttribute("sd_mbr", surfaciqueDistMBR);
							}
						}
					}
				}

				// If the "circle" detection is activated
				if (circleDetection) {
					// If circle
					if (PolygonUtil.isCircle(polygon, iMiller)) {
						newFeature.setAttribute("circle", new Integer(1));
					} else {
						newFeature.setAttribute("circle", new Integer(0));
					}
				}

				// Classification polygon / multipolygon
				if (flagMulti) {
					newFeature.setAttribute("multipolygon", "true");					
				} else {
					newFeature.setAttribute("multipolygon", "false");
				}

				// Add the new feature in the new collection
				newCollection.add(newFeature);
			}
		} else { // If the "circle" detection is activated only
			if (circleDetection) {
				for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
					Feature feature = it.next();
					Feature newFeature = new BasicFeature(fsNew);
					for (int i = 0 ; i < feature.getAttributes().length ; i++) {
						newFeature.setAttribute(i, feature.getAttribute(i));
					}

					boolean flagMulti = false;

					Polygon polygon = null;
					if (feature.getGeometry() instanceof Polygon) {
						polygon = (Polygon)feature.getGeometry();
					} else if (feature.getGeometry() instanceof MultiPolygon) {
						MultiPolygon multipolygon = (MultiPolygon)feature.getGeometry();

						// Choice of the polygon which has the biggest area in the multipolygon
						double areaMax = Double.MIN_VALUE;
						for (int i = 0 ; i < multipolygon.getNumGeometries() ; i++) {
							Polygon p = (Polygon)multipolygon.getGeometryN(i);
							if (p.getArea() > areaMax) {
								polygon = p;
							}
						}
						flagMulti = true;
					}

					// If circle
					if (PolygonUtil.isCircle(polygon, iMiller)) {
						newFeature.setAttribute("circle", new Integer(1));
					} else {
						newFeature.setAttribute("circle", new Integer(0));
					}

					// Classification polygon / multipolygon
					if (flagMulti) {
						newFeature.setAttribute("multipolygon", "true");					
					} else {
						newFeature.setAttribute("multipolygon", "false");
					}

					// Add the new feature in the new collection
					newCollection.add(newFeature);
				}
			}
		}

		return newCollection;
	}

}