/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.plugin.polygon;

import java.util.Iterator;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jump.feature.AttributeType;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.task.TaskMonitor;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.LayerEventType;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.plugin.ThreadedPlugIn;
import com.vividsolutions.jump.workbench.ui.ErrorDialog;
import com.vividsolutions.jump.workbench.ui.GUIUtil;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;
import com.vividsolutions.jump.workbench.ui.plugin.FeatureInstaller;

import fr.ign.cogit.alpage.morphal.i18n.I18NPlug;


/** 
 * Plugin to compute perimeters and areas of {@link Polygon}s or/and {@link MultiPolygon}s
 * from a given {@link Layer}. 
 * 
 * Attention: the {@link Feature}s with a {@link Geometry} different from a {@link Polygon}
 * or a {@link MultiPolygon} are not computed.
 * 
 * @author Eric Grosso - IGN / Laboratoire COGIT
 *
 * @version 0.1 (2011-01-09)
 */
public class PolygonPerimeterAreaPlugin extends AbstractPlugIn implements ThreadedPlugIn {

	private static Logger LOGGER = Logger.getLogger(PolygonPerimeterAreaPlugin.class);
	
	////////////////////////////////////////////////////////////////////////////////////
	// Internationalisation
	
	private final static String MENU = I18NPlug.getI18N("project");
	private final static String MENU_SUBTITLE = I18NPlug.getI18N("polygon");
	private final static String TITLE = I18NPlug.getI18N("polygon.PolygonPerimeterAreaPlugin.title");
	private final static String DESCRIPTION = I18NPlug.getI18N("polygon.PolygonPerimeterAreaPlugin.description");

	private final static String LAYER = I18NPlug.getI18N("polygon.PolygonPerimeterAreaPlugin.layer");
	
	private final static String ACCURACY = I18NPlug.getI18N("polygon.PolygonPerimeterAreaPlugin.accuracy");
	
	private final static String ERROR = I18NPlug.getI18N("error");
	private final static String EMPTY = I18NPlug.getI18N("error.emptyCollection");
	private final static String ERROR_INCORRECT_TYPE_TITLE = I18NPlug.getI18N("error.polygon.incorrectGeometricTypeTitle");
	private final static String ERROR_INCORRECT_TYPE = I18NPlug.getI18N("error.polygon.incorrectGeometricType");
	private final static String ERROR_INCORRECT_TYPE_ALL = I18NPlug.getI18N("error.polygon.incorrectGeometricTypeAll");
	
	////////////////////////////////////////////////////////////////////////////////////
	
	private static MultiInputDialog dialog;
	
	
	/** 
	 * Initialisation.
	 * 
	 * @param context
	 */
	@Override
	public void initialize(PlugInContext context) {
		FeatureInstaller featureInstaller = new FeatureInstaller(context.getWorkbenchContext());
		featureInstaller.addMainMenuItem(
				this,
				new String[] { MENU, MENU_SUBTITLE },
				TITLE,
				false,
				null,
				new MultiEnableCheck()
				.add(context.getCheckFactory().createTaskWindowMustBeActiveCheck())
				.add(context.getCheckFactory().createAtLeastNLayersMustExistCheck(1)));
	}
	
	
	/** 
	 * Global execution.
	 * 
	 * @param context 
	 * @return true if process is executed, otherwise false
	 */
	@Override
	public boolean execute(PlugInContext context) {
		this.reportNothingToUndoYet(context);
		dialog = new MultiInputDialog(context.getWorkbenchFrame(), getName(), true);
		setDialogValues(dialog, context);
		GUIUtil.centreOnWindow(dialog);
		dialog.setVisible(true);
		if (! dialog.wasOKPressed()) { return false; }
		return true;
	}
	

	/** 
	 * Initialisation of the dialog interface.
	 * 
	 * @param dialog
	 * @param context
	 */
	private void setDialogValues(MultiInputDialog dialog, PlugInContext context) {
		dialog.setSideBarDescription(DESCRIPTION);

		dialog.addLayerComboBox(LAYER,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());
		
		dialog.addCheckBox(ACCURACY, false);
	}


	/** 
	 * Execution.
	 * 
	 * @param monitor
	 * @param context
	 */
	public void run(TaskMonitor monitor, PlugInContext context) {
		computePerimeterArea(context);
		System.gc();
	}

	
	/** 
	 * Initialisation of the computation of perimeters and areas.
	 * 
	 * @param context
	 * @return boolean
	 */
	@SuppressWarnings("unchecked")
	private boolean computePerimeterArea(PlugInContext context) {
		
		// Layer to process
		Layer layer = dialog.getLayer(LAYER);

		// Features of the layer to process
		final FeatureCollection collection = layer.getFeatureCollectionWrapper();

		// If the collection is empty
		if (collection.isEmpty()) {
			ErrorDialog.show(context.getWorkbenchFrame(),
					ERROR,
					EMPTY,
					"");
			return false;
		}

		// Management of incorrect geometric input (other than Polygon or MultiPolygon)
		// This preprocess is not long and allows to users to make a choice about what they
		// want to do.
		int nPMP = 0; // Number of features with a Polygon or a MultiPolygon geometric type
		
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature f = it.next();
			if (f.getGeometry() instanceof Polygon || 
					f.getGeometry() instanceof MultiPolygon) {
				nPMP++;
			}
		}
		
		int size = collection.size();
		
		if (nPMP == size) { // All features have a correct type
			computePerimeterArea(collection);
		} else if (nPMP == 0) { // All features have an incorrect type
			ErrorDialog.show(
					context.getWorkbenchFrame(),
					ERROR_INCORRECT_TYPE_TITLE,
					ERROR_INCORRECT_TYPE_ALL,
					"");
			return false;
		} else {
			int result = JOptionPane.showConfirmDialog(
					context.getWorkbenchFrame(),
					(size - nPMP) + " " + ERROR_INCORRECT_TYPE,
					ERROR_INCORRECT_TYPE_TITLE,
					JOptionPane.OK_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE);
			
			// If OK, the process is launched. Otherwise, it is not.
			if (result == JOptionPane.OK_OPTION) {
				computePerimeterArea(collection);
			} else {
				return false;
			}
		}
		return true;
	}
	
	
	/** 
	 * Compute and add fields about perimeters and areas of a {@link Layer}
	 * which contains {@link Polygon}s or/and {@link MultiPolygon}.
	 *
	 * @param collection
	 * @return true
	 */
	@SuppressWarnings("unchecked")
	private boolean computePerimeterArea(FeatureCollection collection) {

		Layer layer = dialog.getLayer(LAYER);
		boolean accuracy = dialog.getCheckBox(ACCURACY).isSelected();
		
		FeatureSchema fsNew = (FeatureSchema) collection.getFeatureSchema().clone();
		fsNew.addAttribute("perimeter", AttributeType.DOUBLE);
		fsNew.addAttribute("area", AttributeType.DOUBLE);
		
		FeatureDataset newCollection = new FeatureDataset(fsNew);

		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature feature = it.next();
			Feature newFeature = new BasicFeature(fsNew);
			for (int i = 0 ; i < feature.getAttributes().length ; i++) {
				newFeature.setAttribute(i, feature.getAttribute(i));
			}
			
			Geometry geometry = feature.getGeometry();
			// Compute only for Polygon and Multipolygon geometric types
			if (geometry instanceof Polygon || geometry instanceof MultiPolygon) {
				if(accuracy) {
					newFeature.setAttribute("perimeter",
							new Double(new Long(Math.round(geometry.getLength() * 100)).doubleValue() / 100.0));
					newFeature.setAttribute("area",
							new Double(new Long(Math.round(geometry.getArea() * 100)).doubleValue() / 100.0));
				} else {
					newFeature.setAttribute("perimeter", new Double(geometry.getLength()));
					newFeature.setAttribute("area", new Double(geometry.getArea()));
				}
				newCollection.add(newFeature);
			} else { // If different from Polygon or MultiPolygon, feature not computed
				LOGGER.warn("Feature ID = " + feature.getID() + " is not process: not a Polygon or a MultiPolygon.");
			}
		}
		
        layer.setFeatureCollection(newCollection);
        layer.fireLayerChanged(LayerEventType.METADATA_CHANGED);

		return true;
	}

}