/* 
 * This file is part of the MorphAL plugin project for OpenJUMP.
 *
 * The MorphAL plugin project aims at providing some tools to analyse specifically
 * historical vector data even if they can be also used to analyse current data.
 * 
 * These tools have been developped during the ALPAGE project and are described
 * in the documentation -- detailed installation instructions are also provided
 * with this documentation --provided with this code or accessible through
 * the following links:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/cartography (English version)
 * http://lamop.univ-paris1.fr/alpage/index.php/cartographie (French version)
 * 
 * For more information, feel free to visit: http://lamop.univ-paris1.fr/alpage
 * 
 * Copyright (C) 2010 ALPAGE
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact the ALPAGE team:
 * http://lamop.univ-paris1.fr/alpage_english/index.php/contact-us
 * 
 */
package fr.ign.cogit.alpage.morphal.plugin.statistics;

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.JComboBox;
import javax.swing.JTextField;

import com.vividsolutions.jump.feature.AttributeType;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.task.TaskMonitor;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.LayerEventType;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.plugin.ThreadedPlugIn;
import com.vividsolutions.jump.workbench.ui.ErrorDialog;
import com.vividsolutions.jump.workbench.ui.GUIUtil;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;
import com.vividsolutions.jump.workbench.ui.plugin.FeatureInstaller;

import fr.ign.cogit.alpage.morphal.i18n.I18NPlug;

/** 
 * Plugin to classify the values of a given numerical attribute,
 * i.e. an attribute with a type "integer" or "double".
 *
 * @author Eric Grosso - IGN / Laboratoire COGIT
 *  
 * @version 0.1 (2011-01-09)
 */
public class ClassificationIntegerDoublePlugin extends AbstractPlugIn implements ThreadedPlugIn {
	
	////////////////////////////////////////////////////////////////////////////////////
	// Internationalisation
		
	private final static String MENU = I18NPlug.getI18N("project");
	private final static String MENU_SUBTITLE = I18NPlug.getI18N("statistics");
	private final static String TITLE = I18NPlug.getI18N("statistics.ClassificationIntegerDoublePlugin.title");
	
	private final static String LAYER = I18NPlug.getI18N("statistics.ClassificationIntegerDoublePlugin.layer");
	private final static String DESCRIPTION = I18NPlug.getI18N("statistics.ClassificationIntegerDoublePlugin.description");

	private final static String ATTRIBUTE = I18NPlug.getI18N("statistics.ClassificationIntegerDoublePlugin.attribute");
	private final static String STEP = I18NPlug.getI18N("statistics.ClassificationIntegerDoublePlugin.classificationStep");

	private final static String ERROR = I18NPlug.getI18N("error");
	private final static String EMPTY = I18NPlug.getI18N("error.emptyCollection");
	
	////////////////////////////////////////////////////////////////////////////////////

	private static MultiInputDialog dialog;
	
	private JComboBox layerCombo = null, itemsCombo;
	private JTextField classificationStep;
	
	
	/** 
	 * Initialisation
	 * 
	 * @param context
	 */
	@Override
	public void initialize(PlugInContext context) {
		FeatureInstaller featureInstaller = new FeatureInstaller(context.getWorkbenchContext());
		featureInstaller.addMainMenuItem(
				this,
				new String[] { MENU, MENU_SUBTITLE },
				TITLE,
				false,
				null,
				new MultiEnableCheck()
				.add(context.getCheckFactory().createTaskWindowMustBeActiveCheck())
				.add(context.getCheckFactory().createAtLeastNLayersMustExistCheck(1)));
	}

	
	/** 
	 * Global execution
	 * 
	 * @param context
	 * @return true if process is executed, otherwise false
	 */
	@Override
	public boolean execute(PlugInContext context) {
		this.reportNothingToUndoYet(context);
		dialog = new MultiInputDialog(context.getWorkbenchFrame(), getName(), true);
		setDialogValues(dialog, context);
		GUIUtil.centreOnWindow(dialog);
		dialog.setVisible(true);
		if (! dialog.wasOKPressed()) { return false; }
		return true;
	}

	
	/** 
	 * Initialisation of the dialog interface
	 * 
	 * @param dialog
	 * @param context
	 */
	private void setDialogValues(MultiInputDialog dialog, PlugInContext context) {
		dialog.setSideBarDescription(DESCRIPTION);
		
		this.layerCombo = dialog.addLayerComboBox(LAYER,
				context.getLayerNamePanel().getLayerManager().getLayer(0),
				context.getLayerNamePanel().getLayerManager());

		// TODO: to improve if a layer does not contain any attribute

		this.layerCombo.addItemListener(new ItemsChange());
		
		// Items initialisation
		Layer layer = context.getLayerNamePanel().getLayerManager().getLayer(0);

		Collection<String> items = new ArrayList<String>();
		// Add items to the collection if their type is INTEGER or DOUBLE
		FeatureSchema fs = layer.getFeatureCollectionWrapper().getFeatureSchema();
		for (int i = 0 ; i < fs.getAttributeCount() ; i++) {
			if (fs.getAttributeType(i).toString().equals("INTEGER")
					|| fs.getAttributeType(i).toString().equals("DOUBLE")) {
				items.add(fs.getAttributeName(i));
			}
		}

		this.itemsCombo = dialog.addComboBox(ATTRIBUTE, "", items, "");
		
		if (items.size() == 0) {
			this.itemsCombo.setEnabled(false);	
		} else {
			this.itemsCombo.setSelectedIndex(0);
			this.itemsCombo.setEnabled(true);
		}
		
		// Classification
		this.classificationStep = dialog.addDoubleField(STEP, 10.0, 5, STEP);
		this.classificationStep.setEnabled(true);
	}
	
	
	/** Internal class to allow the management of items for a specific layer */	
	class ItemsChange implements ItemListener {

		@Override
		public void itemStateChanged(ItemEvent arg0) {
			// Delete items of the old layer
			itemsCombo.removeAllItems();
			
			// Select items with a type INTEGER or DOUBLE
			int numAtt = 0;
			Layer layer = dialog.getLayer(LAYER);
			FeatureSchema fs = layer.getFeatureCollectionWrapper().getFeatureSchema();
			for (int i = 0 ; i < fs.getAttributeCount() ; i++) {
				if (fs.getAttributeType(i).toString().equals("INTEGER")
						|| fs.getAttributeType(i).toString().equals("DOUBLE")) {
					itemsCombo.addItem(fs.getAttributeName(i));
					numAtt++;
				}
			}
			// If the layer does not contain at least one attribute
			// the combo box becomes unavailable
			if (numAtt == 0) {
				itemsCombo.setEnabled(false);	
			} else {
				// By default, selection of the first item
				itemsCombo.setSelectedIndex(0);
				itemsCombo.setEnabled(true);	
			}
		}
	}
	
	
	/** 
	 * Execution
	 * 
	 * @param monitor
	 * @param context
	 */
	public void run(TaskMonitor monitor, PlugInContext context) {
		classification(context);
		System.gc();
	}

	
	/**
	 * Initialisation of the classification computation
	 * 
	 * @param context
	 * @return
	 */
	private boolean classification(PlugInContext context) {
		
		// Layer to process
		Layer layer = dialog.getLayer(LAYER);

		// Collection to process
		FeatureCollection collection = layer.getFeatureCollectionWrapper();

		if (collection.isEmpty()) {
			ErrorDialog.show(context.getWorkbenchFrame(),
					ERROR,
					EMPTY,
					"");
			return false;
		}

		// Classification
		classification(layer, collection);
		
		return true;
	}

	
	/** 
	 * Classification of the values for a given attribute.
	 * If the step is negative or null, it is automatically assigned to 1.0
	 * 
	 * @param layer
	 * @param collection
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private boolean classification(Layer layer, FeatureCollection collection) {

		// Classification step
		double classificationStepValue = dialog.getDouble(STEP);
		if (classificationStepValue <= 0) {
			classificationStepValue = 1.0;
		}
		
		String attributeName = itemsCombo.getSelectedItem().toString();

		FeatureSchema fsClone = (FeatureSchema) collection.getFeatureSchema().clone();
		fsClone.addAttribute("classe", AttributeType.INTEGER);
		
		FeatureDataset collectionNew = new FeatureDataset(fsClone);		
		
		for (Iterator<Feature> it = collection.iterator() ; it.hasNext();) {
			Feature feature = it.next();
			
			Feature featureNew = new BasicFeature(fsClone);
			for (int i = 0 ; i < feature.getAttributes().length ; i++) {
				featureNew.setAttribute(i, feature.getAttribute(i));
			}
			
			int classeNumber = new Double((Double)feature.getAttribute(attributeName) / classificationStepValue).intValue();
			featureNew.setAttribute("classe", classeNumber);

			collectionNew.add(featureNew);			
		}
		
        layer.setFeatureCollection(collectionNew);
        layer.fireLayerChanged(LayerEventType.METADATA_CHANGED);

		return true;
	}

}